/***************************************************************************
*
*   Module:             diags
*
*   Filename:           sysevent.c
*
*   Author:             Brian Tran
*
*   Created:            09/03/2002
*
*   Description:
*
*   This device is read-only from user perspective. The bloking read returns when there is an interrput on system events
*   such as fan faults, over temperature, power supply failure. A user library is needed to find out what exactly what when wrong.
*   This device only returns the one-byte cause code of the event.  
*
****************************************************************************
*                 Source Control System Information
*
*    $Id: 270sc_sysevent.c,v 1.18 2004/05/13 08:34:52 brian Exp $
*
****************************************************************************
*                       Modification History
*
*    $Log: 270sc_sysevent.c,v $
*    Revision 1.18  2004/05/13 08:34:52  brian
*    Bugzilla number:
*    Support for getting reset reason in 270
*
*    Revision 1.17  2004/05/11 07:51:29  brian
*    Bugzilla number:
*    Support for changing i2c speed on the fly
*
*    Revision 1.16  2004/05/06 21:47:08  brian
*    Bugzilla number:
*    Check the masterchange bit first before waiting on a queue
*
*    Revision 1.15  2004/05/06 00:12:26  brian
*    Bugzilla number:
*    Support for mastership change intr in 270
*
*    Revision 1.14  2004/05/05 06:39:54  brian
*    Bugzilla number:
*    Route smm intr from a node to fpga intr in a core
*
*    Revision 1.13  2004/04/23 21:37:57  brian
*    Bugzilla number:
*    Support for i2croute with single mgmt card in any slot
*
*    Revision 1.12  2004/04/20 02:16:26  brian
*    Bugzilla number:
*    Add the reset hang fix for 270
*
*    Revision 1.11  2004/04/02 03:04:00  brian
*    Bugzilla number:
*    Now turn off all intr and let the master turn it on later in system/hw_lib/init.c
*
*    Revision 1.10  2004/04/01 02:40:29  brian
*    Bugzilla number:
*    Disable unused interrupts
*
*    Revision 1.9  2004/03/24 03:56:22  brian
*    Bugzilla number:
*    Initial support for system event handling
*
*    Revision 1.8  2004/01/13 23:53:51  sarah
*    put back previous changes, fix header file includes
*
*    Revision 1.6  2003/12/12 15:56:45  sarah
*    init intrs
*
*    Revision 1.5  2003/11/12 21:25:23  brian
*    Bugzilla number:
*    Fixed a bug in setting the mgmt switches
*
*    Revision 1.4  2003/10/30 23:15:15  brian
*    Bugzilla number:
*    Fixed a typo
*
*    Revision 1.3  2003/10/24 21:36:30  brian
*    Bugzilla number:
*    Add access to anafa seeprom
*
*    Revision 1.2  2003/10/24 02:25:04  brian
*    Bugzilla number:
*    Add support for i2c route setting in 270
*
*    Revision 1.1  2003/10/22 00:25:55  brian
*    Bugzilla number:
*    Initial version of sysevent for 270
*
*
***************************************************************************
*                       (c) COPYRIGHT 2000, 2001, 2002
*                           TopSpin Communications, Inc.
*                           Palo Alto, CA
**************************************************************************/
#ifdef RCSID
static const char rcsid[] = "$Id: 270sc_sysevent.c,v 1.18 2004/05/13 08:34:52 brian Exp $";
#endif


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/machdep.h>
#include <asm/ocp.h>
#include <linux/i2c-errcode.h>
#include <topspin_types.h>
#include <fpga.h>
#include <system_interrupt_code.h>
#include "270sc_sysevent.h"
#include "ds1672_rtc.h"
#include "dcr.h"

extern int global_i2c_debug;
extern int i2c_controller_fast_mode;
extern unsigned char i2c_controller_slave_addr;
extern struct semaphore i2c_lock;


extern int phy_iic_xfer(void *vaddr, unsigned short dev_addr, int access, char *buf, int count, int controller);
extern int phy_iic_init(void *vaddr);

int sysevent_init_module(void);
void sysevent_cleanup_module(void);
int seti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller);


/* Function Prototypes */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset);
static unsigned int sysevent_poll(struct file *file, poll_table *wait);
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);


/* global vars */
unsigned char ts270_my_chassis_id;
unsigned char ts270_my_slot_id;
unsigned char ts270_is_a_node;



/* File Operation Table */
static struct file_operations char_fops = {
    owner: THIS_MODULE,
    read: sysevent_read,
    write: sysevent_write,
    ioctl: sysevent_ioctl,
    poll: sysevent_poll,
};

static wait_queue_head_t sysevent_waitq;
static wait_queue_head_t masterchange_waitq;
static wait_queue_head_t delay_waitq;


static char *ts270_reset_reason;
static volatile int fpga_int_active;
static volatile int smm_fault_int_active;
static volatile int smm_fault_recover_int_active;
static volatile int smm_healthy_int_active;
static volatile int smm_healthy_recover_int_active;
static volatile int hardware_err_sync_int_active;
static volatile int err_fru_id;



static int timeout = 5;

static void *fpga_base_addr;
static void *i2c0_base_addr;
static void *i2c1_base_addr;
static volatile void *gpio_base_addr;

static void deassert_AB_intr(void)
{
    unsigned char data;

    if (ts270_is_a_node == 0 ) {
        return;
    }

    data = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    data &= (~0xc0);

    writeb(data, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));

    return;
}


/* assert_AB_intr
 * --------------
 * This function is for a node to call. No ops for a core.
 * A node calls this function to send intr to the cores (both A and B) to notify
 * the cores of a SMM events. The core wil then use ethernet communication to determine what
 * is going on with the SMM on the node.
 */
static void assert_AB_intr(void)
{
    unsigned char data;

    if (ts270_is_a_node == 0 ) {
        return;
    }

    data = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    data |= 0xc0;

    writeb(data, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    
    return;
}


/* interrupt handler */
static void fpga_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    unsigned char data;

    data = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
    if ((data & 0x01) != 0) {	/* mastership change */
        /* clear it */
        writeb(0x01, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));        
        wake_up_interruptible(&masterchange_waitq);
    }

    fpga_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);
}


static void smm_fault_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    int dcr_val;

    /* check current polarity */
    dcr_val = mfdcr(UIC0_POLARITY_REG);
    if (dcr_val & TS270_SMM_FAULT_INT_MASK) {       /* the bit is 1, rising edge - recovery event */
        smm_fault_recover_int_active = 1;
        deassert_AB_intr();
    }
    else {
        smm_fault_int_active = 1;
        assert_AB_intr();
    }

    /* reverse the polarity to catch the other edge */
    dcr_val ^= TS270_SMM_FAULT_INT_MASK;
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    wake_up_interruptible(&sysevent_waitq);

}

static void smm_healthy_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    int dcr_val;

    /* check current polarity */
    dcr_val = mfdcr(UIC0_POLARITY_REG);
    if (dcr_val & TS270_SMM_HEALTHY_INT_MASK) {       /* the bit is 1, rising edge - recovery event */
        /* smm_healthy_recover_int_active = 1; */
        deassert_AB_intr();
    }
    else {
        smm_healthy_int_active = 1;
        assert_AB_intr();
    }

    /* reverse the polarity to catch the other edge */
    dcr_val ^= TS270_SMM_HEALTHY_INT_MASK;
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    wake_up_interruptible(&sysevent_waitq);

}


int read_reset_reason(char *buf, char **start, off_t offset, int len, int *eof, void *data)
{
    if (offset > 0) {
        return 0;
    }

    len = sprintf(buf, ts270_reset_reason);
    buf[len] = '\n';
    len ++;
    *eof = 1;
    
    return len;
}






int sysevent_init_module(void)
{
    int dcr_val;
    unsigned char data;
    char *sram_sig;

    /* init global vars */
    smm_fault_int_active = 0;
    smm_fault_recover_int_active = 0;
    fpga_int_active = 0;
    smm_healthy_int_active = 0;
    smm_healthy_recover_int_active = 0;
    hardware_err_sync_int_active = 0;



    /* register char device */
    if (register_chrdev(SYSEVENT_MAJOR, "syseventmodule", &char_fops)) {
        printk(KERN_ERR "Unable to register syseventmodule\n");
        return -EIO;
    }

    create_proc_read_entry("resetreason", 0, NULL, read_reset_reason, NULL);

    fpga_base_addr = ioremap_nocache(FPGA_REGADDR_BASE, 0x200); 
    sram_sig = (char *)PPC440_SRAM_VIRT_BASE; 
    i2c0_base_addr = ioremap_nocache(I2C_CONTROLLER0_BASE, 0x12); 
    i2c1_base_addr = ioremap_nocache(I2C_CONTROLLER1_BASE, 0x12); 
    gpio_base_addr = (volatile void *)ioremap_nocache(GPIO_BASE_REGADDR, 0x1000);

    /* inspect the sram for reset signature */
    sram_sig[RESET_SIG_LEN] = 0;

    if (!strcmp(sram_sig, SOFTWARE_RESET)) {
        ts270_reset_reason = SOFTWARE_RESET;
    }
    else if (!strcmp(sram_sig, WATCHDOG_RESET)) {
        ts270_reset_reason = WATCHDOG_RESET;
    }
    else {
        ts270_reset_reason = POWER_ON_RESET;
    }

    strcpy(sram_sig, WATCHDOG_RESET);

    /* make sure gpio 11: tri-state disable, no open drain */
        {
            volatile unsigned int data;
            /* disable tri-state - make sure tri-state bit == 1 */
            data = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_TRISTATE_REGADDR)));
            data |= GPIO_11_MASK | GPIO_9_MASK;
            writel(__cpu_to_le32(data), (volatile void *)((unsigned int)gpio_base_addr + GPIO_TRISTATE_REGADDR));

            /* no open drain - make sure open drain bit == 0 */
            data = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OPENDRAIN_REGADDR)));
            data &= ~(GPIO_11_MASK | GPIO_9_MASK);
            writel(__cpu_to_le32(data), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OPENDRAIN_REGADDR));
        }


    data = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_CHASID_REG_OFFSET));
    ts270_my_chassis_id = (data >> SC_270_FPGA_CHASID_SHIFT) & SC_270_FPGA_CHASID_MASK;
    ts270_my_slot_id = data & SC_270_FPGA_FABRIC_SLOT_ID_MASK;
    if ((ts270_my_slot_id != SC_270_FPGA_CORE1_SLOT_NUM) && (ts270_my_slot_id != SC_270_FPGA_CORE2_SLOT_NUM)) {
        ts270_is_a_node = 1;
    }
    else {
        ts270_is_a_node = 0;
    }



    /* init wait queue */
    init_waitqueue_head(&sysevent_waitq);
    init_waitqueue_head(&masterchange_waitq);
    init_waitqueue_head(&delay_waitq);


    /* configure IRQ4 (Summit Healthy), IRQ5 (FPGA Int), and IRQ6 (Summit Int) to be trigger type, falling edge */
    dcr_val = mfdcr(UIC0_TRIGGER_REG);
    dcr_val |= TS270_UIC0_INT_MASK;
    mtdcr(UIC0_TRIGGER_REG, dcr_val);

    dcr_val = mfdcr(UIC0_POLARITY_REG);
    dcr_val &= (~TS270_UIC0_INT_MASK);
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    /* clear int status */
    dcr_val = mfdcr(UIC0_STATUS_REG);
    dcr_val |= TS270_UIC0_INT_MASK;
    mtdcr(UIC0_STATUS_REG, dcr_val);

    
    /* request interrupt */
    if (request_irq(FPGA_IRQ_NUM, fpga_int_handler, SA_INTERRUPT, "FPGA Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", FPGA_IRQ_NUM);
        return -1;
    }

    if (request_irq(SMM_FAULT_IRQ_NUM, smm_fault_int_handler, SA_INTERRUPT, "Summit Fault Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", SMM_FAULT_IRQ_NUM);
        return -1;
    }

    if (request_irq(SMM_HEALTHY_IRQ_NUM, smm_healthy_int_handler, SA_INTERRUPT, "Summit Healthy Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", SMM_HEALTHY_IRQ_NUM);
        return -1;
    }


    /* init the FPGA. Do this differently for node vs core */
    if (ts270_is_a_node) {
        /* for a node, disable A_pw_fail and B_pw_fail. We do not deal with power fail right now */
        writeb(0x30, (void *) ((int) fpga_base_addr + SC_270_FPGA_PWR_INT_CNTL_REG_OFFSET));
        
        /* clear the A_err and B_err change bits here */
        writeb(0x06, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET));

#if 0
        /* then enable A_err and B_err interrupts */
        writeb(0x30, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
#endif

        /* disable A_err and B_err */
        writeb(0x0, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));

        /* enable master change intr */
        writeb(0x10, (void *) ((int) fpga_base_addr + SC_270_FPGA_MISC_INT_CNTL_REG_OFFSET)); 
    }
    else {	/* a core, same for both master and non-master */
        
        /* Presence change bit */
        writeb(0x01, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_PRES_REG2_OFFSET));
        /* enable presence and master change intr */
        writeb(0x90, (void *) ((int) fpga_base_addr + SC_270_FPGA_MISC_INT_CNTL_REG_OFFSET)); 

        /* Power supplies: clear Change and disable intr */
        writeb(0xf0, (void *) ((int) fpga_base_addr + SC_270_FPGA_PWR_INT_CNTL_REG_OFFSET));
        /* writeb(0x0f, (void *) ((int) fpga_base_addr + SC_270_FPGA_PWR_INT_CNTL_REG_OFFSET)); */

        /* Module and LIM ERRn */
        writeb(0xff, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG0_OFFSET));
        writeb(0xff, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG1_OFFSET));
        /* disable Error Intr from  all*/
        writeb(0x0, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_ERR_INT_EN_REG0_OFFSET));
        writeb(0x0, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_ERR_INT_EN_REG1_OFFSET));

        
        /* Module and LIM INTn */
        writeb(0xff, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_INT_EDGEDET_REG0_OFFSET));
        writeb(0xff, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_INT_EDGEDET_REG1_OFFSET));
        /* disable Intr from all cards */
        writeb(0x0, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_INT_CPUINT_EN_REG0_OFFSET));
        writeb(0x0, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_INT_CPUINT_EN_REG1_OFFSET));

        /* C_ERR */
        writeb(0x01, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET));
        /* writeb(0x08, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET)); */
        /* disable C_err */
        writeb(0x0, (void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    }

    return 0;
}


void sysevent_cleanup_module(void)
{

    unregister_chrdev(SYSEVENT_MAJOR, "syseventmodule");
    remove_proc_entry("resetreason", NULL);
    iounmap(fpga_base_addr);
    iounmap(i2c0_base_addr);
    iounmap(i2c1_base_addr);
    iounmap((void *)gpio_base_addr);
    free_irq(FPGA_IRQ_NUM, NULL);
    free_irq(SMM_FAULT_IRQ_NUM, NULL);
    free_irq(SMM_HEALTHY_IRQ_NUM, NULL);
}



/* Read routine */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset)
{
    int interrupt_code;
    int delay;
    int minor;
    struct inode *inode = file->f_dentry->d_inode;
    unsigned char data;

    minor = MINOR(inode->i_rdev);


#if 0
    unsigned char fpga_errAB_change;
    unsigned char fpga_errC_change;
    unsigned char fpga_err0_change;
    unsigned char fpga_err1_change;
    unsigned char fpga_int0_change;
    unsigned char fpga_int1_change;
    unsigned char fpga_ps_change;
    unsigned char fpga_presence_change;
#endif


    if (minor == SYSEVENT_MASTERCHANGE_MINOR) {
        data = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
        if ((data & 0x01) != 0) {   /* mastership change */
            /* clear it */
            writeb(0x01, (void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
        }
        else {
            interruptible_sleep_on(&masterchange_waitq);
        }
        interrupt_code = MASTERCHANGE_INT_CODE;
        copy_to_user(buf, &interrupt_code, sizeof(int));
        return sizeof(int);
    }
    

    if ((smm_fault_int_active == 0) && (smm_fault_recover_int_active == 0) && (fpga_int_active == 0) && (smm_healthy_int_active == 0) && (hardware_err_sync_int_active == 0) && (smm_healthy_recover_int_active == 0)) {
        interruptible_sleep_on_timeout(&sysevent_waitq, timeout * HZ);
    }

    if (hardware_err_sync_int_active == 1) {
        interrupt_code = HARDWARE_ERR_SYNC_INT_CODE;
        interrupt_code |= err_fru_id << SYSTEM_INT_SHIFT;
        hardware_err_sync_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }


    if (smm_fault_int_active == 1) {
        interrupt_code = SMM_FAULT_INT_CODE;
        smm_fault_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (smm_fault_recover_int_active == 1) {
        interrupt_code = SMM_FAULT_RECOVER_INT_CODE;
        smm_fault_recover_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }


    if (fpga_int_active == 1) {
        interrupt_code = FPGA_INT_CODE;
        fpga_int_active = 0;
        delay = 1;
        goto sysevent_read_return;
    }

    if (smm_healthy_int_active == 1) {
        interrupt_code = SMM_HEALTHY_INT_CODE;
        smm_healthy_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (smm_healthy_recover_int_active == 1) {
        interrupt_code = SMM_HEALTHY_RECOVER_INT_CODE;
        smm_healthy_recover_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

#if 0
    /* check the change again, just in case of race condition */
    if (ts270_is_a_node) {
/* no longer enable AB Error */
#if 0
        fpga_errAB_change = (SC_270_FPGA_A_ERR_CHANGE_MASK | SC_270_FPGA_B_ERR_CHANGE_MASK) & readb((void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET));
#endif
        fpga_errAB_change = 0;
        if (fpga_errAB_change != 0) {
            interrupt_code = FPGA_INT_CODE;
            delay = 1;
            goto sysevent_read_return;
        }
    }
    else {		/* a core */
        /* Presence bit */
        fpga_presence_change = SC_270_FPGA_PRESENCE_CHANGE_MASK & readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_PRES_REG2_OFFSET));
        fpga_ps_change = SC_270_FPGA_PS_CHANGE_MASK & readb((void *) ((int) fpga_base_addr + SC_270_FPGA_PWR_INT_CNTL_REG_OFFSET));
        /* ignore Error change from all */
#if 0
        fpga_err0_change = 0x0f & readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG0_OFFSET));
        fpga_err1_change = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG1_OFFSET));
#endif
        fpga_err0_change = 0;
        fpga_err1_change = 0;

        /* only look at INTr change from mgmt and fan */
        fpga_int0_change = 0x0f & readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_INT_EDGEDET_REG0_OFFSET));
        /* fpga_int1_change = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_INT_EDGEDET_REG1_OFFSET)); */
        fpga_int1_change = 0;
#if 0
        fpga_errC_change = SC_270_FPGA_C_ERR_CHANGE_MASK & readb((void *) ((int) fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET));
#endif
        fpga_errC_change = 0;

        if ((fpga_presence_change != 0) || (fpga_ps_change != 0) || (fpga_err0_change != 0) || (fpga_err1_change != 0) || (fpga_int0_change != 0) || (fpga_int1_change != 0) || (fpga_errC_change != 0)) { 
            interrupt_code = FPGA_INT_CODE;
            delay = 1;
            goto sysevent_read_return;
        }

    }

#endif


    /* return with interrupt code == 0 */
    interrupt_code = 0;
    delay = 0;


sysevent_read_return:

    /* delay before returning */
    interruptible_sleep_on_timeout(&delay_waitq, delay * HZ);
    copy_to_user(buf, &interrupt_code, sizeof(int));
    return sizeof(int);

}


/* Write routine
 * -------------
 * Users want to clear a bit mask in the interrupt code. The value should have the bit zero, all other bits are 1
 */
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset)
{
    int value;
    int controller;

    copy_from_user(&value, buf, sizeof(int));

    if ((value & I2C_DEBUG_CODE) == I2C_DEBUG_CODE) {
        global_i2c_debug = value & 0xff;
    }

    if ((value & I2C_CLOCK_CODE) == I2C_CLOCK_CODE) {
        i2c_controller_fast_mode = (value >> SYSEVENTWRITE_I2CCLOCK_MODE_SHIFT) & SYSEVENTWRITE_I2CCLOCK_MODE_MASK;
        i2c_controller_slave_addr = (value >> SYSEVENTWRITE_I2CCLOCK_SLAVE_SHIFT) & SYSEVENTWRITE_I2CCLOCK_SLAVE_MASK;
        controller = (value >> SYSEVENTWRITE_I2CCLOCK_BUS_SHIFT) & SYSEVENTWRITE_I2CCLOCK_BUS_MASK;
        down(&i2c_lock);
        if (controller == 0) {
            phy_iic_init(i2c0_base_addr);
        }
        else {
            phy_iic_init(i2c1_base_addr);
        } 
        up(&i2c_lock);
    }

    if ((value & FRU_ERR_CODE) == FRU_ERR_CODE) {
        hardware_err_sync_int_active = 1;
        err_fru_id = value & 0xffff;
        wake_up_interruptible(&sysevent_waitq);
    }


    return sizeof(int);
}

/* Ioctl routine */
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    return 0;
}


static unsigned int sysevent_poll(struct file *file, poll_table *wait)
{
    return 0;
}










/* deifne the mux select value to set i2c route
 * This mux select is a one-byte value:
 * - bit 0: 1 for on-board, 0 for off-board
 * - bit 1: select mgmt card. 0 for mgmt card 1, 1 for mgmt card 2.
 * If the mgmt switch value is NOT to select on-board mgmt card devices, this bit is used to determine which mgmt card to go thru.
 * 0 means default route, 1 means alternate route.
 * If there is only one mgmt card present, this bit is a don'tcare.
 * If there are 2 mgmt cards present, default route is: core 1 goes thru mgmt card 1, core 2 goes thru mgmt card 2.
 * Alternate route is the opposite of the default route. Alternate route is used by upper level application for error handling: when the default route fails.
 * - bit [3:2]: lim number - 0 for node card, 1 for min 1, 2 for lim 2
 * - bit [7:4]: management switch value.
 */

int seti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller)
{
    unsigned char data;
    unsigned char presence;
    unsigned char mgmt_mux_addr;
    unsigned char card_mux_addr;
    unsigned char mgmt_switch_addr;
    unsigned char other_mgmt_switch_addr;
    int need_to_change_mux;
    int need_to_set_node_mux = 0;
    unsigned char mgmtswitch;
    unsigned char lim_num;
    int ret;
    unsigned char buf[4];

    if (muxselect == 0xff) {
        return 0;
    }

    if (controller == 1) {		/* on 270, controller 1 is NOT connected to the i2c switch */
        return 0;
    }

    mgmtswitch = (muxselect >> 4) & 0x0f;   /* bits 7:4 */

    if ((muxselect & 0x1) == 0x1) {     /* bit 0 == 1 for local on-board devices */
        if (mgmtswitch == 14) {			/* anafa2 seeprom 1 */
            data = 8;					/* channel 3 */
        }
        else if (mgmtswitch == 15) {	/* anafa2 seeprom 2 */
            data = 16;					/* channel 4 */
        }
        else {							/* summit, etc. */
            data = 4;					/* channel 2 */
        }
        if ((ret = iic_xfer(vaddr, (0xff << 8) | LOCAL_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
            return IIC_ERR_LOCALSWITCH_ERR;
        }
        else {
            return 0;
        }
    }

    /* if here, we are going to/thru a mgmt card. Only core cards can go thru mgmt card */
    if (ts270_is_a_node) {
        return 0;
    }

  if (mgmtswitch == 6) {		/* we are targeting devices on mgmt card. In this case, use bit 1 to determine which card */

    if ((muxselect & 0x2) == 0x2) {     /* bit 1 = 1 for mgmt card 1 */
        mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
        data = 2;	/* channel 1 */
    }
    else {								/* bit 1 = 0 for mgmt card 0 */
        mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
        data = 1;	/* channel 0 */
    }

  }
  else {	/* we are using mgmt card as a mean to reach node/lim cards. Inthis case, use bit 1 to determine ewither default route or alternate route */
    /* first see if mgmt cards are present */
   presence = readb((void *) ((int) fpga_base_addr + SC_270_FPGA_MOD_PRES_REG2_OFFSET));
   if ((presence & 0x0C) == 0x0C) {   /* both mgmtA & B present */ 
    /* then you have choices */ 
    if ((muxselect & 0x2) == 0x2) {     /* bit 1 = 1 for alternate route */
        if (ts270_my_slot_id == SC_270_FPGA_CORE1_SLOT_NUM) {
            mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
            data = 2;   /* channel 1 */
        }
        else {
            mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
            data = 1;   /* channel 0 */
        } 
    }
    else {								/* bit 1 = 0 for default route */
        if (ts270_my_slot_id == SC_270_FPGA_CORE1_SLOT_NUM) {
            mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
            data = 1;   /* channel 0 */
        }
        else {
            mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
            data = 2;   /* channel 1 */
        }
    }
   }		
   else {	/* not both mgmt cards are present */
        if ((presence & 0x08) != 0) {   /* mgmtA present */
            mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
            data = 1;
        }
        else if ((presence & 0x04) != 0) {  /* mgmtB present */
            mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
            data = 2;   /* channel 1 */
        }
        else {  /* no one present */
            return IIC_ERR_MGMT0_ERR;	/* just pick one error code for mgmt cards not presence */ 
        }
   }	/* end if of checking mgmt card presence */
  }		/* end if of using mgmt cards to reach other devices */

    if ((ret = iic_xfer(vaddr, (0xff << 8) | LOCAL_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_LOCALSWITCH_ERR;
    }

    /* now read the mux to see the control bits  */
    data = MGMT_MUX_CONTROL_REG_OFFSET;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, READACCESS, &data, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }


    need_to_change_mux = 0;
    /* here is the format of the control: bit 0 = mygrab, bit 1 = Nmygrab, bit 2 = myonoff, bit 3 = Nmyonoff */
    /* make sure the channel is on */
    if ((data & 0x0c) == 0x0c) {
        need_to_change_mux = 1;
        data &= (~0x04);
    }

    if ((data & 0x0c) == 0) {
        need_to_change_mux = 1;
        data |= 0x04;
    }

    /* make sure I can grab the channel */
    if ((data & 0x03) == 0x01) {
        need_to_change_mux = 1;
        data &= (~0x01);
    }

    if ((data & 0x03) == 0x02) {
        need_to_change_mux = 1;
        data |= 0x01;
    }

    if (need_to_change_mux) {
        buf[0] = MGMT_MUX_CONTROL_REG_OFFSET;
        buf[1] = data;
        if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, WRITEACCESS, buf, 2, controller)) != 2) {
            if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
                return IIC_ERR_MGMT0_ERR;
            }
            else {
                return IIC_ERR_MGMT1_ERR;
            }
        }
    }


    /* now set the switch on the mgmt card */
    if (mgmtswitch < 4) {   /* try to talk to node cards */
        mgmt_switch_addr = MGMT_SWITCH0_I2C_ADDR;
        other_mgmt_switch_addr = MGMT_SWITCH1_I2C_ADDR;
        need_to_set_node_mux = 1;
    }
    else {
        mgmt_switch_addr = MGMT_SWITCH1_I2C_ADDR;
        other_mgmt_switch_addr = MGMT_SWITCH0_I2C_ADDR;
        mgmtswitch -= 4;
        /* no mux for on-board mgmt card (channel 2) and for power supplies (channel 5 & 6) */
        if ((mgmtswitch == 2) || (mgmtswitch == 5) || (mgmtswitch == 6)) {
            need_to_set_node_mux = 0;
        }
        else {
            need_to_set_node_mux = 1;
        }
    }

    /* deselect the other switch */
    data = 0;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | other_mgmt_switch_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    /* then select this switch */
    data = 1 << mgmtswitch;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_switch_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    /* if not trying to talk to node cards or the core cards, we are done */
    if (need_to_set_node_mux == 0) {
        return 0;
    }

    /* if here, we are trying to get to the node or the core cards */
    lim_num = (muxselect >> 2) & 0x03;		/* bits 3:2 */
    if (lim_num == 0) {
        card_mux_addr = NODE_MUX_I2C_ADDR;
    }
    else if (lim_num == 1) {
        card_mux_addr = LIM1_MUX_I2C_ADDR;
    }
    else {
        card_mux_addr = LIM2_MUX_I2C_ADDR;
    }


    /* now read the mux to see the control bits  */
    data = MGMT_MUX_CONTROL_REG_OFFSET;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | card_mux_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_TARGETCARD_MUX_ERR;
    }

    if ((ret = iic_xfer(vaddr, (0xff << 8) | card_mux_addr, READACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_TARGETCARD_MUX_ERR;
    }

    need_to_change_mux = 0;
    /* here is the format of the control: bit 0 = mygrab, bit 1 = Nmygrab, bit 2 = myonoff, bit 3 = Nmyonoff */
    /* make sure the channel is on */
    if ((data & 0x0c) == 0x0c) {
        need_to_change_mux = 1;
        data &= (~0x04);
    }

    if ((data & 0x0c) == 0) {
        need_to_change_mux = 1;
        data |= 0x04;
    }

    /* make sure I can grab the channel */
    if ((data & 0x03) == 0x01) {
        need_to_change_mux = 1;
        data &= (~0x01);
    }

    if ((data & 0x03) == 0x02) {
        need_to_change_mux = 1;
        data |= 0x01;
    }

    if (need_to_change_mux) {
        buf[0] = MGMT_MUX_CONTROL_REG_OFFSET;
        buf[1] = data;
        if ((ret = iic_xfer(vaddr, (0xff << 8) | card_mux_addr, WRITEACCESS, buf, 2, controller)) != 2) {
            return IIC_ERR_TARGETCARD_MUX_ERR;
        }
    }

    /* final switch for LIMs */
    if (lim_num == 0) {
        return 0;
    }

    data = 1;       /* all LIM on-board devices reside on channel 0 for now */
    if (lim_num == 1) {
        if ((ret = iic_xfer(vaddr, (0xff << 8) | LIM1_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
            return IIC_ERR_LIM_SWITCH_ERR;
        }
    }
    else {
        if ((ret = iic_xfer(vaddr, (0xff << 8) | LIM2_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
            return IIC_ERR_LIM_SWITCH_ERR;
        }
    }

    return 0;
}


/* init_ts_watchdog
 * ------------
 * Start the hardware watchdog.
 */
void init_ts_watchdog(void)
{
    volatile unsigned int gpio_output;

    gpio_output = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR)));

    /* write 1 */
    gpio_output |= GPIO_11_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    /* write 0 */
    gpio_output &= ~GPIO_11_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    return;
}


/* ts270_anafa_reset
 * ------------
 */
void ts270_anafa_reset(void)
{
    volatile unsigned int gpio_output;

    gpio_output = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR)));

    /* write 0 */
    gpio_output &= ~GPIO_9_MASK;
    gpio_output &= ~GPIO_10_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    return;
}


/* ts270_disable_dma
 * -----------------
 */
void ts270_disable_dma(void)
{
    writeb(0x3, (void *) ((int) fpga_base_addr + SC_270_FPGA_DMA_CONTROL_OFFSET));
}




