/***************************************************************************
*
*   Module:             diags
*
*   Filename:           sysevent.c
*
*   Author:             Brian Tran
*
*   Created:            09/03/2002
*
*   Description:
*
*   This device is read-only from user perspective. The bloking read returns when there is an interrput on system events
*   such as fan faults, over temperature, power supply failure. A user library is needed to find out what exactly what when wrong.
*   This device only returns the one-byte cause code of the event.  
*
****************************************************************************
*                 Source Control System Information
*
*    $Id: cc_sysevent.c,v 1.3 2003/12/15 18:47:05 brian Exp $
*
****************************************************************************
*                       Modification History
*
*    $Log: cc_sysevent.c,v $
*    Revision 1.3  2003/12/15 18:47:05  brian
*    Bugzilla number:
*    Added debug messages
*
*    Revision 1.2  2003/08/01 06:20:52  brian
*    Bugzilla number: 622
*    Make sysevent wait queue available for i2c_serial_console driver in case of LM sensor interrupt
*
*    Revision 1.1  2003/07/25 00:25:37  brian
*    Bugzilla number:
*    I2C console support for IGR_CC
*
*    Revision 1.6  2003/05/08 21:00:07  brian
*    Bugzilla number:
*    Added delay before returing. Added more interrupt types
*
*    Revision 1.5  2003/05/05 23:46:55  scott
*    merge from 1.1.2 branch
*
*    Revision 1.4.2.1  2003/05/03 06:20:09  brian
*    Bugzilla number:
*    Add async event support for switch cards
*
*    Revision 1.4  2003/04/30 18:59:33  brian
*    Bugzilla number:
*    Added support for async event handling in IGR CC
*
*    Revision 1.3  2003/04/24 00:27:06  brian
*    Bugzilla number: 920
*    Removed the while loop here. Let the user loop instead.
*
*    Revision 1.2  2003/03/11 02:23:31  brian
*    Check for power supply fan fail and power supply not present at startup
*
*    Revision 1.1  2003/03/06 05:47:57  brian
*    Kernel module to handle system interrupts (fan, sensors, etc.) Initial version
*
*
***************************************************************************
*                       (c) COPYRIGHT 2000, 2001, 2002
*                           TopSpin Communications, Inc.
*                           Palo Alto, CA
**************************************************************************/
#ifdef RCSID
static const char rcsid[] = "$Id: cc_sysevent.c,v 1.3 2003/12/15 18:47:05 brian Exp $";
#endif


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/machdep.h>
#include <asm/ocp.h>
#include "cc_sysevent.h"
#include "dcr.h"


extern int cc_i2c_serial_int_handler(int slot);
extern int global_i2c_debug;

int sysevent_init_module(void);
void sysevent_cleanup_module(void);



/* Function Prototypes */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset);
static unsigned int sysevent_poll(struct file *file, poll_table *wait);
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);


/* File Operation Table */
static struct file_operations char_fops = {
    owner: THIS_MODULE,
    read: sysevent_read,
    write: sysevent_write,
    ioctl: sysevent_ioctl,
    poll: sysevent_poll,
};

wait_queue_head_t sysevent_waitq;
static wait_queue_head_t delay_waitq;

#define MAX_SLOT	13

static wait_queue_head_t i2ctty_waitq[MAX_SLOT];
static want_to_handle_ibml[MAX_SLOT];
static volatile int fpga_int_active;
static volatile int lm_int_active;


static int timeout = 15;

static void *fpga_base_addr;


static void fpga_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    volatile unsigned char fpga_int1;
    volatile unsigned char fpga_int2;
    int i;
    unsigned char mask;
    int use_default_handler = 1;

    fpga_int1 = readb((void *) ((int) fpga_base_addr + FPGA_INT1_REGADDR));

    if (fpga_int1 != 0) {
        for (i = 8, mask = 0x1; i >= 1; i --, mask <<= 1) {
            if ((fpga_int1 & mask) == 0) {
                continue;
            }

            /* we got an interrupt here */
            fpga_int_active = 1;
            /* printk("Goto slot %d handler\n", i); */

            if (want_to_handle_ibml[i] == 1) {
                wake_up_interruptible(&i2ctty_waitq[i]);
                use_default_handler = 0;
            }
            else if (cc_i2c_serial_int_handler(i) == 0) {
                use_default_handler = 0;
            }
            else {
                use_default_handler = 1;
            }
        }
    }

    fpga_int2 = readb((void *) ((int) fpga_base_addr + FPGA_INT2_REGADDR));
    fpga_int2 &= 0xf0;	/* mask out part of int2 for swtich and an interrupts */

    if (fpga_int2 != 0) {
        for (i = 4, mask = 0x10; i >= 1; i --, mask <<= 1) {
            if ((fpga_int2 & mask) == 0) {
                continue;
            }

            /* we got an interrupt here */
            fpga_int_active = 1;
            /* printk("Goto slot %d handler\n", i + 8); */

            if (want_to_handle_ibml[i + 8] == 1) {
                wake_up_interruptible(&i2ctty_waitq[i + 8]);
                use_default_handler = 0;
            }
            else if (cc_i2c_serial_int_handler(i + 8) == 0) {
                use_default_handler = 0;
            }
            else {
                use_default_handler = 1;
            }

        }

    }

    if (use_default_handler == 1) {
        fpga_int_active = 1;
        wake_up_interruptible(&sysevent_waitq);
    }

    return;

}

static void lm_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    lm_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);
}


int sysevent_init_module(void)
{
    int dcr_val;
    unsigned char fpga_status;
    int i;

    /* init global vars */
    fpga_int_active = 0;
    lm_int_active = 0;


    /* register char device */
    if (register_chrdev(SYSEVENT_MAJOR, "syseventmodule", &char_fops)) {
        printk(KERN_ERR "Unable to register syseventmodule\n");
        return -EIO;
    }

    /* get virt addr of gpio input register */
    fpga_base_addr = ioremap_nocache(FPGA_REGADDR_BASE, 0x10);

    /* init wait queue */
    init_waitqueue_head(&sysevent_waitq);
    init_waitqueue_head(&delay_waitq);

    for (i = 0; i < MAX_SLOT; i ++) {
        init_waitqueue_head(&i2ctty_waitq[i]);
        want_to_handle_ibml[i] = 0;
    }

    /* configure IRQ3, 4 to be trigger type, falling edge */
    dcr_val = mfdcr(UIC0_TRIGGER_REG);
    dcr_val |= 0x30;
    mtdcr(UIC0_TRIGGER_REG, dcr_val);

    dcr_val = mfdcr(UIC0_POLARITY_REG);
    dcr_val &= (~0x30);
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    /* clear int status */
    dcr_val = mfdcr(UIC0_STATUS_REG);
    dcr_val |= 0x30;
    mtdcr(UIC0_STATUS_REG, dcr_val);

 
    /* request interrupt */
    if (request_irq(FPGA_IRQ_NUM, fpga_int_handler, SA_INTERRUPT, "FPGA Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", FPGA_IRQ_NUM);
        return -1;
    }


    if (request_irq(LM_IRQ_NUM, lm_int_handler, SA_INTERRUPT, "LM Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", LM_IRQ_NUM);
        return -1;
    }

    /* clear fpga change register so that any new interrupt will be seen */
    writeb(0xff, (void *) ((int) fpga_base_addr + FPGA_CHANGE1_REGADDR));
    writeb(0xff, (void *) ((int) fpga_base_addr + FPGA_CHANGE2_REGADDR));
    writeb(0xff, (void *) ((int) fpga_base_addr + FPGA_CHANGE3_REGADDR));

    /* write 1 to the mask interrupt register to enable interrupt */
    writeb(0xff, (void *) ((int) fpga_base_addr + FPGA_MASK1_REGADDR));
    writeb(0xff, (void *) ((int) fpga_base_addr + FPGA_MASK2_REGADDR));

    fpga_status = readb((void *) ((int) fpga_base_addr + FPGA_STATUS2_REGADDR));
    if (((fpga_status & FPGA_PS1_PRESENT_MASK) == 0) || ((fpga_status & FPGA_PS2_PRESENT_MASK) == 0)) {   /* ps not present */
        fpga_int_active = 1;
    }

    if (((fpga_status & FPGA_PS1_AC_FAIL_MASK) == 0) || ((fpga_status & FPGA_PS2_AC_FAIL_MASK) == 0)) {   /* ps ac fail */
        fpga_int_active = 1;
    }


    fpga_status = readb((void *) ((int) fpga_base_addr + FPGA_STATUS4_REGADDR));

    if (((fpga_status & FPGA_FAN1_PRESENT_MASK) == 0) || ((fpga_status & FPGA_FAN2_PRESENT_MASK) == 0)) {   /* fan not present */
        fpga_int_active = 1;
    }



    return 0;
}


void sysevent_cleanup_module(void)
{

    unregister_chrdev(SYSEVENT_MAJOR, "syseventmodule");
    iounmap(fpga_base_addr);
    free_irq(FPGA_IRQ_NUM, NULL);
    free_irq(LM_IRQ_NUM, NULL);

#ifdef WANT_TO_HANDLE_PS_FAIL
    free_irq(PS_FAIL_IRQ_NUM, NULL);
#endif WANT_TO_HANDLE_PS_FAIL 

}

#define IGR_CC_POLLING_NUM		4
const int polling_code[IGR_CC_POLLING_NUM] = {EXT_INT_CODE, SWITCH_INT_CODE, SLAVE_CC_INT_CODE, FAN_INT_CODE};


/* Read routine */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset)
{
    int interrupt_code;
    volatile unsigned char fpga_change1;
    volatile unsigned char fpga_change2;
    volatile unsigned char fpga_change3;
    volatile unsigned char fpga_int1;
    volatile unsigned char fpga_int2;
    static int i = 0;
    int delay;

    if ((lm_int_active == 0) && (fpga_int_active == 0)) {
        interruptible_sleep_on_timeout(&sysevent_waitq, timeout * HZ);
    }

    if (lm_int_active == 1) {
        interrupt_code = LM_INT_CODE;
        lm_int_active = 0;
        goto cc_sysevent_read_return;
    }

    if (fpga_int_active == 1) {
        interrupt_code = FPGA_INT_CODE;
        fpga_int_active = 0;
        goto cc_sysevent_read_return;
    }

    fpga_change1 = readb((void *) ((int) fpga_base_addr + FPGA_CHANGE1_REGADDR));
    fpga_change2 = readb((void *) ((int) fpga_base_addr + FPGA_CHANGE2_REGADDR));
    fpga_change3 = readb((void *) ((int) fpga_base_addr + FPGA_CHANGE3_REGADDR));
    fpga_int1 = readb((void *) ((int) fpga_base_addr + FPGA_INT1_REGADDR));
    fpga_int2 = readb((void *) ((int) fpga_base_addr + FPGA_INT2_REGADDR));


    /* FIXME - fpga continously generates intr on power supply AC changes. Remove fpga_change3 for now */
    /* if ((fpga_change1 != 0) || (fpga_change2 != 0) || (fpga_change3 != 0) || (fpga_int1 != 0) || (fpga_int2 != 0)) { */
    if ((fpga_change1 != 0) || (fpga_change2 != 0) || (fpga_int1 != 0) || (fpga_int2 != 0)) {
        interrupt_code = FPGA_INT_CODE;
        goto cc_sysevent_read_return;
    }

    /* return with interrupt code == polling code */
    interrupt_code = polling_code[i];

    i ++;
    if (i == IGR_CC_POLLING_NUM) {
        i = 0;
    }

cc_sysevent_read_return:

    /* delay before returning */
    delay = 1;
    interruptible_sleep_on_timeout(&delay_waitq, delay * HZ);
    copy_to_user(buf, &interrupt_code, sizeof(int));
    return sizeof(int);
    
}


static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset)
{
    int value;
    int slot;

    copy_from_user(&value, buf, sizeof(int));

    if ((value & I2C_DEBUG_CODE) == I2C_DEBUG_CODE) {
        global_i2c_debug = value & 0xff;
        return sizeof(int);
    }

    /* first byte is the slot */
    slot = value & 0xff;

    /* second byte indicates if apps wants to handle ibml intr for this slot */
    want_to_handle_ibml[slot] = (value >> 8) & 0xff; 

    fpga_int_active = 0;

    return sizeof(int);

}

/* Ioctl routine */
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{

    /* treat cmd as slot number 1 - 12 */
    if (fpga_int_active == 0) {
        interruptible_sleep_on(&i2ctty_waitq[cmd]);
    }

    if (fpga_int_active == 1) {
        fpga_int_active = 0;
    }
    return 0;
}


static unsigned int sysevent_poll(struct file *file, poll_table *wait)
{
    unsigned int mask = 0;
    int minor;
    struct inode *inode = file->f_dentry->d_inode;

    minor = MINOR(inode->i_rdev); 

    poll_wait(file, &i2ctty_waitq[minor], wait);

    if (fpga_int_active == 1) {
        mask |= POLLIN;
    }

    return mask;
}




