/*
 * I2C serial driver
 *
 */

#include <linux/config.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/interrupt.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/machdep.h>
#include <asm/ocp.h>

#include "ds1672_rtc.h"

#define DRIVER_VERSION "v1.1"
#define DRIVER_AUTHOR "Brian Tran"
#define DRIVER_DESC "I2C TTY driver"

/* Module information */
MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

#define I2C_SERIAL_RATE	32
#define DELAY_TIME		(HZ /I2C_SERIAL_RATE) 

#define TINY_TTY_MAJOR		249	/* experimental range */
#define TINY_TTY_MINORS		1

#define PIC_PHY_ADDR				0x1f0200000LL
#define I2C_CONTROLLER1_PHY_ADDR    0x140000500LL
#define PIC_PSP_IB_INTR_HI          0x60
#define PIC_PSP_IB_INTR_LO          0x61


extern wait_queue_head_t slaveread_complete_iic_wait[2];
extern wait_queue_head_t slaveread_service_iic_wait[2];
extern wait_queue_head_t slavewrite_complete_iic_wait[2];
extern wait_queue_head_t slavewrite_service_iic_wait[2];



int i2c_serial_slave_write_handler(unsigned char slave_status);
int i2c_serial_slave_read_handler(unsigned char slave_status);


static void gateway_i2c_serial_callback(void *arg);

#define MAX_WRITE_SIZE						256

static unsigned char write_buffer[MAX_WRITE_SIZE];
static unsigned char send_buffer[MAX_WRITE_SIZE + 3];
static volatile int current_write_pointer;
static volatile int current_read_pointer;
static volatile int bytes_in_buffer;
static volatile int write_done;
static struct semaphore master_i2c_serial_write_lock;
static wait_queue_head_t wait_until_sent_waitq;
static wait_queue_head_t wait_until_read_waitq;
static wait_queue_head_t write_buffer_overflow_waitq;

static struct tq_struct gateway_i2c_serial_callback_tq;


struct tiny_serial {
	struct tty_struct	*tty;		/* pointer to the tty for this device */
    int                 open_count; /* number of times this port has been opened */
    struct semaphore    sem;        /* locks this structure */
	struct timer_list	*timer;
};

static void *pic_base_addr;
static void *i2c_base_addr;


static int			tiny_refcount;
static struct tty_driver	tiny_tty_driver;
static struct tty_struct	*tiny_tty[TINY_TTY_MINORS];
static struct termios		*tiny_termios[TINY_TTY_MINORS];
static struct termios		*tiny_termios_locked[TINY_TTY_MINORS];
static struct tiny_serial	*tiny_table[TINY_TTY_MINORS];	/* initially all NULL */


static void tiny_timer (unsigned long data)
{
	struct tiny_serial *tiny = (struct tiny_serial *)data;
	struct tty_struct *tty;
    int i;


	if (!tiny) {
		return;
    }

	tty = tiny->tty;

    if (!write_done) {
        /* printk("Not write_done\n"); */ 
        goto tiny_timer_exit;
    }

    if (current_write_pointer == 0) {	/* nothing to write */
        /* printk("Empty write buffer\n"); */ 
        goto tiny_timer_exit;
    }

    /* put the count at the beginning of the send buffer */
    /* the first 2 bytes willbe the count, LSB first */
    send_buffer[0] = current_write_pointer & 0xff;
    send_buffer[1] = (current_write_pointer >> 8) & 0xff;

    /* copy write buffer to send buffer */
    for (i = 0; i < current_write_pointer; i ++) {
        send_buffer[i + 2] = write_buffer[i];
    }

    gateway_i2c_serial_callback_tq.data = (void *)(current_write_pointer + 2);
    bytes_in_buffer = (current_write_pointer + 2);
    write_done = 0;
    current_write_pointer = 0;
    current_read_pointer = 0;

    /* printk("Schedule to write\n"); */
    schedule_task(&gateway_i2c_serial_callback_tq);

tiny_timer_exit:
	
	/* resubmit the timer again */
	tiny->timer->expires = jiffies + DELAY_TIME;
	add_timer (tiny->timer);
}


int i2c_serial_slave_read_handler(unsigned char slave_status)
{
    unsigned char mode_control;
    struct tiny_serial *tiny = tiny_table[0];

    if (!tiny) {
        return -1;
    }


    if ((slave_status & 0x80) != 0) {     /* slave read complete */
        /* make sure to flush the slave buffer */
        mode_control = readb((void *) ((int) i2c_base_addr + I2C_CONTROLLER_MODE_CONTROL));
        writeb(mode_control | 0x80, (void *)((int)i2c_base_addr + I2C_CONTROLLER_MODE_CONTROL));

        /* simply acknowledge the intr */
        writeb(0x80, (void *)((int)i2c_base_addr + I2C_CONTROLLER_EXT_CONTROL));
    }


    if ((slave_status & 0x40) != 0) {     /* slave read need service */
      if (write_done == 0) {
        /* provide 1 byte data to the slave buffer */
        writeb(send_buffer[current_read_pointer], (void *)((int)i2c_base_addr + I2C_CONTROLLER_SLAVE_DBUF));
        current_read_pointer ++;
        bytes_in_buffer --;
        if (current_read_pointer == (MAX_WRITE_SIZE + 3)) {
            printk(KERN_DEBUG "I2C Serial: Read buffer overflow\n");
            current_read_pointer = 0;
        } 
      }
      else {
          /* printk("Provide default data\n"); */ 
          writeb(0, (void *)((int)i2c_base_addr + I2C_CONTROLLER_SLAVE_DBUF));
      }      

      writeb(0x40, (void *)((int)i2c_base_addr + I2C_CONTROLLER_EXT_CONTROL));
    }

    return 0;


}


int i2c_serial_slave_write_handler(unsigned char slave_status)
{
    struct tiny_serial *tiny = tiny_table[0];
    struct tty_struct *tty;
    unsigned char count;
    volatile unsigned char received_byte;
    int i;

    if (!tiny) {
        /* printk("I2C Serial not ready. Default handler\n"); */  
        return -1;
    }


    /* Slave Write Need Service */
    if ((slave_status & 0x10) != 0) {
        /* make sure to empty the slave buf */
        for (i = 0; i < 4; i ++) {
            received_byte = readb((void *) ((int) i2c_base_addr + I2C_CONTROLLER_SLAVE_DBUF));
        }
        /* ack intr */
        writeb(0x10, (void *)((int)i2c_base_addr + I2C_CONTROLLER_EXT_CONTROL));
        return 0;
    }


    tty = tiny->tty;

    if (tty->flip.count >= TTY_FLIPBUF_SIZE)
        tty_flip_buffer_push(tty);


    /* get the number of bytes received */
    count = readb((void *) ((int) i2c_base_addr + I2C_CONTROLLER_TRANSFER_COUNT));
    count = (count >> 4) & 0x07;


    if (count == 4) {	/* we do not expect 4-byte write */
        goto i2c_serial_slave_write_handler_error_exit;
    }

#if 0
    /* special case: read completion signal */
    if (count == 3) {
        wake_up_interruptible(&(slavewrite_service_iic_wait[1]));
        goto i2c_serial_slave_write_handler_error_exit;
    }
#endif


    /* this doesn't actually push the data through unless tty->low_latency is set */
    /* count = readb((void *) ((int) i2c_base_addr + I2C_CONTROLLER_SLAVE_DBUF)); */
/* printk("Got %d bytes\n", count); */
    for (i = 0; i < count; i ++) {
        received_byte = readb((void *) ((int) i2c_base_addr + I2C_CONTROLLER_SLAVE_DBUF));
        tty_insert_flip_char(tty, received_byte, 0);
    }

    tty_flip_buffer_push(tty);
    tty_schedule_flip (tty);

i2c_serial_slave_write_handler_error_exit:

    /* make sure to empty the slave buf */
    received_byte = readb((void *) ((int) i2c_base_addr + I2C_CONTROLLER_MODE_CONTROL));
    writeb(received_byte | 0x80, (void *)((int)i2c_base_addr + I2C_CONTROLLER_MODE_CONTROL));

    /* ack intr */
    writeb(0x20, (void *)((int)i2c_base_addr + I2C_CONTROLLER_EXT_CONTROL));


    return 0;

}



static int tiny_open (struct tty_struct *tty, struct file * filp)
{
	struct tiny_serial *tiny;
	struct timer_list *timer;

	MOD_INC_USE_COUNT;


    current_read_pointer = 0;
    current_write_pointer = 0;
    write_done = 1;
    init_waitqueue_head(&wait_until_sent_waitq);
    init_waitqueue_head(&wait_until_read_waitq);
    init_waitqueue_head(&write_buffer_overflow_waitq);

	/* initialize the pointer incase something fails */
	tty->driver_data = NULL;
    tty->low_latency = 1;

	/* get the serial object associated with this tty pointer */
	tiny = tiny_table[minor(tty->device)];
	if (tiny == NULL) {
		/* first time accessing this device, let's create it */
        gateway_i2c_serial_callback_tq.routine = (void (*)(void *))gateway_i2c_serial_callback;

        init_MUTEX(&master_i2c_serial_write_lock);
 
		tiny = kmalloc (sizeof (*tiny), GFP_KERNEL);
		if (!tiny) {
			MOD_DEC_USE_COUNT;
			return -ENOMEM;
		}

        init_MUTEX (&tiny->sem);
        tiny->open_count = 0;
		tiny->timer = NULL;

		tiny_table[minor(tty->device)] = tiny;
	}

    down (&tiny->sem);

	/* save our structure within the tty structure */
	tty->driver_data = tiny;
	tiny->tty = tty;

    ++tiny->open_count;
    if (tiny->open_count == 1) {
		/* do any hardware initialization needed here */
        /* deassert IBML intr, if any */
        writeb(PIC_PSP_IB_INTR_HI, pic_base_addr);

		/* create our timer and submit it */
		if (!tiny->timer) {
			timer = kmalloc (sizeof (*timer), GFP_KERNEL);
			if (!timer) {
                up (&tiny->sem);
				MOD_DEC_USE_COUNT;
				return -ENOMEM;
			}
			tiny->timer = timer;
		}

		init_timer (tiny->timer);
		tiny->timer->data = (unsigned long )tiny;
		tiny->timer->expires = jiffies + DELAY_TIME;
		tiny->timer->function = tiny_timer;
		add_timer (tiny->timer);
    }

    up (&tiny->sem);

	return 0;
}

static void do_close (struct tiny_serial *tiny)
{
    if (!tiny->open_count) {
        /* port was never opened */
        return;
    }

    down (&tiny->sem);

    --tiny->open_count;
    if (tiny->open_count <= 0) {

        /* shut down our timer and free the memory */
        del_timer (tiny->timer);
    }

 
    MOD_DEC_USE_COUNT;

    up (&tiny->sem);

}

static void tiny_close (struct tty_struct *tty, struct file * filp)
{
	struct tiny_serial *tiny = tty->driver_data;

	if (!tiny) {
		return;
	}

	do_close (tiny);
}	

static int tiny_write (struct tty_struct * tty, int from_user, const unsigned char *buf, int count)
{
	struct tiny_serial *tiny = tty->driver_data;
    int i;
    unsigned long flags;
    unsigned char mybuf[MAX_WRITE_SIZE];


	if (!tiny) {
		return -ENODEV;
	}

/* printk("Request to Write count = %d\n", count); */

    if (count == 0) {
        return 0;
    }

    if (count > MAX_WRITE_SIZE) {
        count = MAX_WRITE_SIZE;
    }


    if (from_user) {
        /* printk("From user\n"); */
        copy_from_user(mybuf, buf, count);
    }
    else {
        /* printk("From kernel\n"); */
        for (i = 0; i < count; i ++) {
            mybuf[i] = buf[i];
        }
    }

    if (in_interrupt()) {
        /* printk("In interrupt time\n"); */
        if ((current_write_pointer + count) > MAX_WRITE_SIZE) {
            /* printk("Write buffer overflow\n"); */
            current_write_pointer = 0;
        }
        for (i = 0; i < count; i ++) {
            write_buffer[current_write_pointer + i] = mybuf[i];
        }
        current_write_pointer += count;
    }
    else {
        while ((current_write_pointer + count) > MAX_WRITE_SIZE) {
            interruptible_sleep_on_timeout(&write_buffer_overflow_waitq, DELAY_TIME);
        }

        save_flags(flags);
        cli(); 
        if ((current_write_pointer + count) > MAX_WRITE_SIZE) {
            /* printk("Write buffer overflow\n"); */
            current_write_pointer = 0;
        }
        for (i = 0; i < count; i ++) {
            write_buffer[current_write_pointer + i] = mybuf[i];
        }
        current_write_pointer += count;
        /* printk("Write pointer = %d\n", current_write_pointer); */
        /* sti(); */
        restore_flags(flags);
    } 


	return count;
}


static void gateway_i2c_serial_callback(void *arg)
{
    int count;
    int timeout = 0;

    count = (int) arg;

/* printk("Need to write %d bytes\n", count); */

    down (&master_i2c_serial_write_lock);

    /* printk("Generate IBML intr\n"); */
    writeb(PIC_PSP_IB_INTR_LO, pic_base_addr);

#if 1

    /* wait for the host to complete reading */  
    /* interruptible_sleep_on_timeout(&(slavewrite_service_iic_wait[1]), DELAY_TIME * 16); */
    while (current_read_pointer != count) {
        /* printk("Current_read = %d\n", current_read_pointer); */
        interruptible_sleep_on_timeout(&wait_until_read_waitq, DELAY_TIME);
        timeout ++;
        if (timeout == I2C_SERIAL_RATE) {
            bytes_in_buffer = 0;
            writeb(PIC_PSP_IB_INTR_HI, pic_base_addr);
            printk(KERN_DEBUG "I2C Serial: Time out\n");
            break;
        }
    }

#endif


/* printk("Did transfer %d bytes\n", current_read_pointer); */

    write_done = 1;

    up (&master_i2c_serial_write_lock);

    return;
}



static int tiny_write_room (struct tty_struct *tty) 
{
	struct tiny_serial *tiny = tty->driver_data;

	if (!tiny) {
		return -ENODEV;
	}

#if 0
    return (MAX_WRITE_SIZE - current_write_pointer - 2); 
#endif

    return 4;

}

static int tiny_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
{
#if TERMIOS_OFF
    struct termios tmp_termios;
#endif

    /* printk("In ioctl, cmd = %d\n", cmd); */
    switch (cmd) {
        case TCGETS:
        case TCGETA:
            if (kernel_termios_to_user_termios((struct termios *)arg, tty->termios)) {
                return -EFAULT;
            }
            return 0;
            break;
#if TERMIOS_OFF
        case TCSETA:
        case TCSETS:
        case TCSETSW:
            if (user_termios_to_kernel_termios(&tmp_termios, (struct termios *) arg)) {
                return -EFAULT;
            }
            *tty->termios = tmp_termios;
            return 0;
            break;
#endif

        default:
            return -ENOIOCTLCMD;
    }
}

static int tiny_chars_in_buffer(struct tty_struct *tty)
{
#if 0
    int real_chars_in_buffer;

    real_chars_in_buffer = bytes_in_buffer + current_write_pointer;

    printk("chars_in_buffer = %d\n", real_chars_in_buffer);
    return (real_chars_in_buffer);
#endif

    return 0;
}

static void tiny_set_termios(struct tty_struct *tty, struct termios *old)
{
    /* printk("In set_termios\n"); */
}

static void tiny_throttle(struct tty_struct *tty)
{   
    /* printk("In throttle\n"); */
}
    
static void tiny_unthrottle(struct tty_struct *tty)
{   
    /* printk("In unthrottle\n"); */
}   
    
static void tiny_stop(struct tty_struct *tty)
{   
    /* printk("In stop\n"); */
}   

static void tiny_start(struct tty_struct *tty)
{   
    /* printk("In start\n"); */
}   

static void tiny_hangup(struct tty_struct *tty)
{   
    /* printk("In hangup\n"); */
}

static void tiny_break_ctl(struct tty_struct *tty, int state)
{
    /* printk("In break_ctl\n"); */
}

static void tiny_flush_buffer(struct tty_struct *tty)
{
}

static void tiny_set_ldisc(struct tty_struct *tty)
{
    /* printk("In set_ldisc\n"); */
}

static void tiny_wait_until_sent(struct tty_struct *tty, int timeout)
{
    /* printk("In wait_until_sent\n"); */
    while ((current_write_pointer != 0) || (bytes_in_buffer != 0)) {
        interruptible_sleep_on_timeout(&wait_until_sent_waitq, DELAY_TIME);
    }
    /* printk("Done wait until sent\n"); */
        
}

static void tiny_send_xchar(struct tty_struct *tty, char ch)
{
    /* printk("In send_xchar\n"); */
}





static struct tty_driver tiny_tty_driver = {
	magic:			TTY_DRIVER_MAGIC,
	driver_name:	"i2ctty",
	name:			"i2ctty",
	major:			TINY_TTY_MAJOR,
	num:			TINY_TTY_MINORS,
	type:			TTY_DRIVER_TYPE_SERIAL,
	subtype:		SERIAL_TYPE_NORMAL,
	flags:			TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS | TTY_DRIVER_RESET_TERMIOS,
	
	refcount:		&tiny_refcount,
	table:			tiny_tty,
	termios:		tiny_termios,
	termios_locked:		tiny_termios_locked,

	open:			tiny_open,
	close:			tiny_close,
	write:			tiny_write,

	write_room:		tiny_write_room,
    chars_in_buffer:    tiny_chars_in_buffer,
    ioctl:			tiny_ioctl,
    set_termios:    tiny_set_termios,
    throttle:       tiny_throttle,
    unthrottle:     tiny_unthrottle,
    stop:           tiny_stop,
    start:          tiny_start,
    hangup:         tiny_hangup,
    break_ctl:      tiny_break_ctl,
    flush_buffer:   tiny_flush_buffer,
    set_ldisc:      tiny_set_ldisc,
    wait_until_sent:    tiny_wait_until_sent,
    send_xchar:     tiny_send_xchar,
};

int __init i2c_serial_init(void)
{
	/* register the tty driver */
	tiny_tty_driver.init_termios          = tty_std_termios;
    tiny_tty_driver.init_termios.c_cflag |= CLOCAL;
    tiny_tty_driver.init_termios.c_lflag &= (~ICANON);

	if (tty_register_driver (&tiny_tty_driver)) {
		printk (KERN_ERR "failed to register I2C tty driver\n");
		return -1;
	}

	printk (KERN_INFO DRIVER_DESC " " DRIVER_VERSION "\n");

    pic_base_addr = ioremap_nocache(PIC_PHY_ADDR, 1);
    i2c_base_addr = ioremap_nocache(I2C_CONTROLLER1_PHY_ADDR, 0x12);

	return 0;
}

void __exit i2c_serial_exit(void)
{
	struct tiny_serial *tiny;
	int i;

	tty_unregister_driver(&tiny_tty_driver);
    iounmap(pic_base_addr);
    iounmap(i2c_base_addr);

	/* shut down all of the timers and free the memory */
	for (i = 0; i < TINY_TTY_MINORS; ++i) {
		tiny = tiny_table[i];
		if (tiny) {
            while (tiny->open_count) {
		        do_close (tiny);
            }
            kfree (tiny->timer);
            kfree (tiny);
            tiny_table[0] = NULL;

        } 
	}
}

