/***************************************************************************
*
*   Module:             diags
*
*   Filename:           sysevent.c
*
*   Author:             Brian Tran
*
*   Created:            09/03/2002
*
*   Description:
*
*   This device is read-only from user perspective. The bloking read returns when there is an interrput on system events
*   such as fan faults, over temperature, power supply failure. A user library is needed to find out what exactly what when wrong.
*   This device only returns the one-byte cause code of the event.  
*
****************************************************************************
*                 Source Control System Information
*
*    $Id: lt_sysevent.c,v 1.3 2003/12/15 18:56:50 brian Exp $
*
****************************************************************************
*                       Modification History
*
*    $Log: lt_sysevent.c,v $
*    Revision 1.3  2003/12/15 18:56:50  brian
*    Bugzilla number:
*    Added hook to turn on/off i2c debug messages
*
*    Revision 1.2  2003/09/08 17:20:37  brian
*    Bugzilla number:
*    Remove extra module_init
*
*    Revision 1.1  2003/07/25 00:28:25  brian
*    Bugzilla number:
*    I2C console support for IGR_LT
*
*    Revision 1.6  2003/05/08 21:00:07  brian
*    Bugzilla number:
*    Added delay before returing. Added more interrupt types
*
*    Revision 1.5  2003/05/05 23:46:55  scott
*    merge from 1.1.2 branch
*
*    Revision 1.4.2.1  2003/05/03 06:20:09  brian
*    Bugzilla number:
*    Add async event support for switch cards
*
*    Revision 1.4  2003/04/30 18:59:33  brian
*    Bugzilla number:
*    Added support for async event handling in IGR CC
*
*    Revision 1.3  2003/04/24 00:27:06  brian
*    Bugzilla number: 920
*    Removed the while loop here. Let the user loop instead.
*
*    Revision 1.2  2003/03/11 02:23:31  brian
*    Check for power supply fan fail and power supply not present at startup
*
*    Revision 1.1  2003/03/06 05:47:57  brian
*    Kernel module to handle system interrupts (fan, sensors, etc.) Initial version
*
*
***************************************************************************
*                       (c) COPYRIGHT 2000, 2001, 2002
*                           TopSpin Communications, Inc.
*                           Palo Alto, CA
**************************************************************************/
#ifdef RCSID
static const char rcsid[] = "$Id: lt_sysevent.c,v 1.3 2003/12/15 18:56:50 brian Exp $";
#endif


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/machdep.h>
#include <asm/ocp.h>
#include "lt_sysevent.h"
#include "dcr.h"

extern int cc_i2c_serial_int_handler(int slot);
extern int global_i2c_debug;

int sysevent_init_module(void);
void sysevent_cleanup_module(void);


/* Function Prototypes */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset);
static unsigned int sysevent_poll(struct file *file, poll_table *wait);
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);



/* File Operation Table */
static struct file_operations char_fops = {
    owner: THIS_MODULE,
    read: sysevent_read,
    write: sysevent_write,
    ioctl: sysevent_ioctl,
    poll: sysevent_poll,
};

static wait_queue_head_t sysevent_waitq;
static wait_queue_head_t delay_waitq;

#define MAX_SLOT    2

static wait_queue_head_t i2ctty_waitq[MAX_SLOT];
static want_to_handle_ibml[MAX_SLOT];


static volatile int ibml_int_active;
static volatile int fpga_int_active;
static volatile int lm_int_active;

static volatile int ibml_int_enable;
static volatile int fpga_int_enable;
static volatile int lm_int_enable;

static volatile int loop_enable;

static int timeout = 5;

static void *gpio_input_addr;
static void *fpga_base_addr;


/* interrupt handler */
#ifdef WANT_TO_HANDLE_PS_FAIL
static void ps_fail_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    wake_up_interruptible(&sysevent_waitq);
}
#endif WANT_TO_HANDLE_PS_FAIL


static void fpga_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    fpga_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);
}

static void ibml_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    int use_default_handler = 1;


    ibml_int_active = 1;

    if (want_to_handle_ibml[1] == 1) {
        wake_up_interruptible(&i2ctty_waitq[1]);
        use_default_handler = 0;
    }
    else if (cc_i2c_serial_int_handler(1) == 0) {
        use_default_handler = 0;
    }
    else {
        use_default_handler = 1;
    }

    if (use_default_handler == 1) {
        wake_up_interruptible(&sysevent_waitq);
    }

}


int sysevent_init_module(void)
{
    int dcr_val;
    unsigned char fpga_status;
    int i;

    /* init global vars */
    ibml_int_active = 0; 
    fpga_int_active = 0; 
    lm_int_active = 0; 

    ibml_int_enable = 1;
    fpga_int_enable = 1;
    lm_int_enable = 1;

    loop_enable = 1;


    /* register char device */
    if (register_chrdev(SYSEVENT_MAJOR, "syseventmodule", &char_fops)) {
        printk(KERN_ERR "Unable to register syseventmodule\n");
        return -EIO;
    }

    /* get virt addr of gpio input register */
    gpio_input_addr = ioremap_nocache(GPIO_REGADDR_BASE, 0x1000); 
    fpga_base_addr = ioremap_nocache(FPGA_REGADDR_BASE, 0x10); 

    /* init wait queue */
    init_waitqueue_head(&sysevent_waitq);
    init_waitqueue_head(&delay_waitq);

    for (i = 0; i < MAX_SLOT; i ++) {
        init_waitqueue_head(&i2ctty_waitq[i]);
        want_to_handle_ibml[i] = 0;
    }


    /* configure IRQ4, 5, 6 to be trigger type, falling edge */
    dcr_val = mfdcr(UIC0_TRIGGER_REG);
    dcr_val |= 0x1C;
    mtdcr(UIC0_TRIGGER_REG, dcr_val);

    dcr_val = mfdcr(UIC0_POLARITY_REG);
    dcr_val &= (~0x1C);
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    /* clear int status */
    dcr_val = mfdcr(UIC0_STATUS_REG);
    dcr_val |= 0x1c;
    mtdcr(UIC0_STATUS_REG, dcr_val);

    /* configure UART1_RTS_DTR to be GPIO15 to read  int from on-board LM80 */
    dcr_val = mfdcr(CPC0_GPIO_REG);
    dcr_val |= 0x00010000;			/* bit 15 */
    mtdcr(CPC0_GPIO_REG, dcr_val);

    
    /* request interrupt */
#ifdef WANT_TO_HANDLE_PS_FAIL
    if (request_irq(PS_FAIL_IRQ_NUM, ps_fail_handler, SA_INTERRUPT, "PS Fail Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", PS_FAIL_IRQ_NUM);
        return -1;
    }
#endif WANT_TO_HANDLE_PS_FAIL

    if (request_irq(FPGA_IRQ_NUM, fpga_int_handler, SA_INTERRUPT, "FPGA Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", FPGA_IRQ_NUM);
        return -1;
    }


    if (request_irq(IBML_IRQ_NUM, ibml_int_handler, SA_INTERRUPT, "IBML Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", IBML_IRQ_NUM);
        return -1;
    }

    /* clear fpga change register so that any new interrupt will be seen */
    writeb(0xff, (void *) ((int) fpga_base_addr + FPGA_CHANGE_REGADDR));

    /* in case something bad has occurred before this routine is called, let's read fpga status register 1 and 2 */
    fpga_status = readb((void *) ((int) fpga_base_addr + FPGA_STATUS1_REGADDR));
    if (((fpga_status & FPGA_PS_AC_FAIL_MASK) == 0) ||
        ((fpga_status & FPGA_PS_FAIL_MASK) == 0) ||
        ((fpga_status & FPGA_PS_FAN_FAIL_MASK) == 0) ||
        ((fpga_status & FPGA_FAN_PRESENT_MASK) != 0) ||
        ((fpga_status & FPGA_FAN_FAIL_MASK) == 0)) {
        fpga_int_active = 1;
    }

    if (((fpga_status & FPGA_PS_TYPE1_PRESENT_MASK) != 0) && ((fpga_status & FPGA_PS_TYPE2_PRESENT_MASK) != 0)) {	/* ps 1 not present */
        fpga_int_active = 1;
    }

    fpga_status = readb((void *) ((int) fpga_base_addr + FPGA_STATUS2_REGADDR));
    if (((fpga_status & FPGA_PS_AC_FAIL_MASK) == 0) ||
        ((fpga_status & FPGA_PS_FAIL_MASK) == 0) ||
        ((fpga_status & FPGA_MEZZANINE_PRESENT_MASK) != 0) ||
        ((fpga_status & FPGA_PS_FAN_FAIL_MASK) == 0)) {
        fpga_int_active = 1;
    }

    if (((fpga_status & FPGA_PS_TYPE1_PRESENT_MASK) != 0) && ((fpga_status & FPGA_PS_TYPE2_PRESENT_MASK) != 0)) {   /* ps 2 not present */
        fpga_int_active = 1;
    }


        
    return 0;
}


void sysevent_cleanup_module(void)
{

    unregister_chrdev(SYSEVENT_MAJOR, "syseventmodule");
    iounmap(gpio_input_addr);
    iounmap(fpga_base_addr);
    free_irq(FPGA_IRQ_NUM, NULL);
    free_irq(IBML_IRQ_NUM, NULL);
#ifdef WANT_TO_HANDLE_PS_FAIL
    free_irq(PS_FAIL_IRQ_NUM, NULL);
#endif WANT_TO_HANDLE_PS_FAIL
}



/* Read routine */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset)
{
    int interrupt_code;
    volatile unsigned int gpio_input;
    unsigned char fpga_change;
    int delay;


    if ((ibml_int_active == 0) && (fpga_int_active == 0)) {
        interruptible_sleep_on_timeout(&sysevent_waitq, timeout * HZ);
    }

    if (ibml_int_active == 1) {	
        interrupt_code = IBML_INT_CODE;
        ibml_int_active = 0;
        delay = 3;
        goto lt_sysevent_read_return;
    }

    if (fpga_int_active == 1) {     
        interrupt_code = FPGA_INT_CODE;
        fpga_int_active = 0;
        delay = 1;
        goto lt_sysevent_read_return;
    }

    fpga_change = readb((void *) ((int) fpga_base_addr + FPGA_CHANGE_REGADDR));
    if (fpga_change != 0) {
        interrupt_code = FPGA_INT_CODE;
        delay = 1;
        goto lt_sysevent_read_return;
    }


    /* time out case. Poll the on-board LM interrupt - GPIO 15 */
    if (lm_int_enable == 1) {

        gpio_input = readl((void *)((unsigned int)gpio_input_addr + GPIO_INPUT_REGADDR));	
        gpio_input &= 0x00000100;

        if (gpio_input == 0) {
            lm_int_enable = 0;
            interrupt_code = LM_INT_CODE;
            delay = 1;
            goto lt_sysevent_read_return;
        }

    }	/* end if lm_int_enable */
 

    /* return with interrupt code == 0 */
    interrupt_code = 0;
    delay = 0;


lt_sysevent_read_return:

    /* delay before returning */
    interruptible_sleep_on_timeout(&delay_waitq, delay * HZ);
    copy_to_user(buf, &interrupt_code, sizeof(int));
    return sizeof(int);

}


/* Write routine
 * -------------
 * Users want to clear a bit mask in the interrupt code. The value should have the bit zero, all other bits are 1
 */
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset)
{
    int cmd;
    int dcr_val;

    copy_from_user(&cmd, buf, sizeof(int));

    if ((cmd & I2C_DEBUG_CODE) == I2C_DEBUG_CODE) {
        global_i2c_debug = cmd & 0xff;
        return sizeof(int);
    }

    switch (cmd) {
        case IBML_INT_CODE:
                                ibml_int_active = 0;
                                break;
        case FPGA_INT_CODE:
                                fpga_int_active = 0;
                                break;
        case LM_INT_CODE:
                                lm_int_active = 0;
                                break;
        case IBML_INT_EN:
                                ibml_int_enable = 1;
                                dcr_val = mfdcr(UIC0_STATUS_REG);
                                dcr_val |= 0x10;
                                mtdcr(UIC0_STATUS_REG, dcr_val);
                                enable_irq(IBML_IRQ_NUM);
                                break;
        case FPGA_INT_EN:
                                fpga_int_enable = 1;
                                dcr_val = mfdcr(UIC0_STATUS_REG);
                                dcr_val |= 0x08;
                                mtdcr(UIC0_STATUS_REG, dcr_val);
                                enable_irq(FPGA_IRQ_NUM);
                                break;
        case LM_INT_EN:
                                lm_int_enable = 1;
                                break;
        case LOOP_ENABLE:
                                loop_enable = 1;
                                break;
        case IBML_INT_DIS:
                                ibml_int_enable = 0;
                                disable_irq(IBML_IRQ_NUM);
                                break;
        case FPGA_INT_DIS:
                                fpga_int_enable = 0;
                                disable_irq(FPGA_IRQ_NUM);
                                break;
        case LM_INT_DIS:
                                lm_int_enable = 0;
                                break;
        case LOOP_DISABLE:
                                loop_enable = 0;
                                break;
        case I2CTTY_ENABLE:
                                want_to_handle_ibml[1] = 1;
                                ibml_int_active = 0;
                                break;
        case I2CTTY_DISABLE:
                                want_to_handle_ibml[1] = 0;
                                ibml_int_active = 0;
                                break;

        default:                break;
    }


    return sizeof(int);

}

/* Ioctl routine */
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{


    /* treat cmd as slot number */
    if (ibml_int_active == 0) {
        interruptible_sleep_on(&i2ctty_waitq[1]);
    }

    if (ibml_int_active == 1) {
        ibml_int_active = 0;
    }
    return 0;
}


static unsigned int sysevent_poll(struct file *file, poll_table *wait)
{
    unsigned int mask = 0;
    int minor;
    struct inode *inode = file->f_dentry->d_inode;

    minor = MINOR(inode->i_rdev);

    poll_wait(file, &i2ctty_waitq[minor], wait);

    if (ibml_int_active == 1) {
        mask |= POLLIN;
    }

    return mask;
}


