/*
   -------------------------------------------------------------------------
   i2c-adap-ibm_ocp.c i2c-hw access for the IIC peripheral on the IBM PPC 405
   -------------------------------------------------------------------------
  
   Current Maintainers:
   Ian DaSilva and Armin Kuster, MontaVista Software, Inc.
   idasilva@mvista.com , akuster@mvista.com or source@mvista.com

   Copyright 2000-2002 MontaVista Software Inc.

   Changes made to support the IIC peripheral on the IBM PPC 405 

   ----------------------------------------------------------------------------
   This file was highly leveraged from i2c-elektor.c, which was created
   by Simon G. Vogl and Hans Berglund:

 
     Copyright (C) 1995-97 Simon G. Vogl
                   1998-99 Hans Berglund

   With some changes from Kysti Mlkki <kmalkki@cc.hut.fi> and even
   Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   ----------------------------------------------------------------------------

   History: 01/20/12 - Armin
   	akuster@mvista.com
   	ported up to 2.4.16+	

   Version 02/03/25 - Armin
       converted to ocp format
       removed commented out or #if 0 code

   Version 1.3: 04/25/02
   	coverted c++ stlye comments to c
	added IIC_DEV macro
	the i2c_adapter->algo_data is a pointer to the iic_algo_i2c_data struct
	the i2c_adapter->data is a pointer to the ocp_struct for the i2c device
	general cleanups

    Version 1.4: 04/30/02 
	added IIC_PORT_DNFS
    Version 1.5: 05/02/02
	fixed mem leak added ocp_dev_free if ocp_reg failed

    Version: 1.6 - Armin
       no longer need extern const paddr structs

    Version: 1.8 - Armin
       name change of ocp_get_dev
       name change for *_driver to *_dev
    Version: 1.9 - Armin 
       changed irq_resource to just irq

   TODO: add PM hooks
    
*/

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <asm/ocp.h>

/*
 * This next section is configurable, and it is used to set the number
 * of i2c controllers in the system.  The default number of instances is 1,
 * however, this should be changed to reflect your system's configuration.
 */

#ifdef MODULE
static int i2c_debug = 0;
#endif

static wait_queue_head_t iic_wait[2];
wait_queue_head_t slaveread_complete_iic_wait[2];
wait_queue_head_t slaveread_service_iic_wait[2];
wait_queue_head_t slavewrite_complete_iic_wait[2];
wait_queue_head_t slavewrite_service_iic_wait[2];
wait_queue_head_t iic_ibmocp_bus_waitq;
static int iic_pending = 0;

extern void *fpga_base_addr;
#define FPGA_PHY_BASE_ADDR		0x1f0100000

extern int i2c_iic_add_bus(struct i2c_adapter *adap);
extern int i2c_iic_del_bus(struct i2c_adapter *adap);
extern int i2c_serial_slave_read_handler(unsigned char slave_status);
extern int i2c_serial_slave_write_handler(unsigned char slave_status);

extern int i2c_serial_in_use;
extern int global_i2c_debug;

/* ----- global defines -----------------------------------------------	*/
#define DEB(x)	if (global_i2c_debug>=1) x
#define DEB2(x) if (global_i2c_debug>=2) x
#define DEB3(x) if (global_i2c_debug>=3) x
#define DEBE(x)	x		/* error messages                               */

#define IIC_DEV(x)	((struct ocp_dev *)adap->data)->x

/* ----- local functions ----------------------------------------------	*/

/*
 * Description:  Put this process to sleep.  We will wake up when the
 * IIC controller interrupts.
 */
void
iic_ibmocp_waitforpin(void *data)
{

	int timeout = 2;
	struct ocp_dev *iic_dev;
    unsigned long flags;

	iic_dev = (struct ocp_dev *) data;

	/*
	 * If interrupts are enabled (which they are), then put the process to
	 * sleep.  This process will be awakened by two events -- either the
	 * the IIC peripheral interrupts or the timeout expires. 
	 */
	if (iic_dev->irq > 0) {
        save_flags(flags);
		cli();
		if (iic_pending == 0) {
			interruptible_sleep_on_timeout(&
						       (iic_wait[iic_dev->num]),
						       timeout * HZ);
		} else
			iic_pending = 0;

        restore_flags(flags);
		/* sti(); */
	} else {
		/*
		 * If interrupts are not enabled then delay for a reasonable amount
		 * of time and return.  We expect that by time we return to the calling
		 * function that the IIC has finished our requested transaction and
		 * the status bit reflects this.
		 * 
		 * udelay is probably not the best choice for this since it is
		 * the equivalent of a busy wait
		 */
		udelay(100);
	}
}

void
simple_iic_ibmocp_waitforpin(int dev_num)
{
    unsigned long flags;

    save_flags(flags);
    cli();
    if (iic_pending == 0) {
        DEB2(printk("Sleep on transfer complete interrupt\n"));
        interruptible_sleep_on_timeout(&(iic_wait[dev_num]), HZ / 200);
        DEB2(printk("Woke up on transfer complete interrupt or timeout\n"));
    }
    else {
        DEB2(printk("Transfer complete interrupt has occurred\n"));
        iic_pending = 0;
    }

    restore_flags(flags);
    /* sti(); */
}


static void i2c_slave_read_handler(struct ocp_dev *iic_dev, unsigned char status)
{
    struct iic_regs *iic;
    unsigned char mode_control;

    if (i2c_serial_slave_read_handler(status) == 0) {
        return;
    } 

    iic = (struct iic_regs *) iic_dev->vaddr;

    /* default handler */
    if ((status & 0x80) != 0) {		/* slave read complete */
        /* make sure to flush the slave buffer */
        mode_control = readb((int) &(iic->mdcntl));  
        writeb(mode_control | 0x80, (int) &(iic->mdcntl));

        /* simply acknowledge the intr */
        writeb(0x80, (int) &(iic->xtcntlss));
    }

    if ((status & 0x40) != 0) {     /* slave read need service */
        /* provide 1 byte data of 0x00 to the slave buffer */
        writeb(0x00, (int) &(iic->sbbuf));
        writeb(0x40, (int) &(iic->xtcntlss));
    }

    return;
}


static void i2c_slave_write_handler(struct ocp_dev *iic_dev, unsigned char status)
{
    struct iic_regs *iic;
    unsigned char mode_control;
    volatile char dummy;
    int i;

    if (i2c_serial_slave_write_handler(status) == 0) {
        return;
    }

    iic = (struct iic_regs *) iic_dev->vaddr;


    /* default handler */
    if ((status & 0x20) != 0) {     /* slave write complete */
        /* make sure to flush the slave buffer */
        mode_control = readb((int) &(iic->mdcntl));  
        writeb(mode_control | 0x80, (int) &(iic->mdcntl));

        /* simply acknowledge the intr */
        writeb(0x20, (int) &(iic->xtcntlss));
    }

    if ((status & 0x10) != 0) {     /* slave write need service */
        /* consume 4 bytes out of the slave buffer */
        for (i = 0; i < 4; i ++) {
            dummy = readb((int) &(iic->sbbuf));
        }
        writeb(0x10, (int) &(iic->xtcntlss));
    }

    return;
}




/*
 * Description: The registered interrupt handler
 */
static void
iic_ibmocp_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
	unsigned char ret;
	struct iic_regs *iic;
	struct ocp_dev *iic_dev;

	iic_dev = (struct ocp_dev *) dev_id;
	iic = (struct iic_regs *) iic_dev->vaddr;

	DEB2(printk("iic_ibmocp_handler: in interrupt handler\n")); 
	/* Read status register */
    ret = readb((int) &(iic->xtcntlss));
    DEB2(printk("iic_ibmocp_handler: ext control status = %x\n", ret)); 
    
    if ((ret & 0xf0) != 0) {		/* slave operation interrupts */
        if ((ret & 0xc0) != 0) {
            i2c_slave_read_handler(iic_dev, ret);
        }
        else {
            i2c_slave_write_handler(iic_dev, ret);
        }
    }
    else {
        /* Master Transfer */
        iic_pending = 1;
        wake_up_interruptible(&(iic_wait[iic_dev->num]));
        DEB2(printk("Trnasfer Complete interrupt. Waking up waiy queue\n"));
    }

    writeb(0x0a, (int) &(iic->sts));

}

/*
 * Description: Release irq and memory
 */
static void
iic_ibmocp_release(void)
{
	int i;
	struct ocp_dev *iic_drv;

	for (i = 0; i < ocp_get_max(IIC); i++) {
		iic_drv = ocp_get_dev(IIC, i);
		if (iic_drv->irq > 0) {
			disable_irq(iic_drv->irq);
			free_irq(iic_drv->irq, 0);
		}
	}

}

/*
 * Description: Does nothing
 */
static int
iic_ibmocp_reg(struct i2c_client *client)
{
	return 0;
}

/*
 * Description: Does nothing
 */
static int
iic_ibmocp_unreg(struct i2c_client *client)
{
	return 0;
}

/*
 * Description: If this compiled as a module, then increment the count
 */
static void
iic_ibmocp_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

/*
 * Description: If this is a module, then decrement the count
 */
static void
iic_ibmocp_dec_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

/*
 * Description: Called when the module is loaded.  This function starts the
 * cascade of calls up through the heirarchy of i2c modules (i.e. up to the
 *  algorithm layer and into to the core layer)
 */
static int __init
iic_ibmocp_init(void)
{
	struct ocp_dev *iic_drv;
	struct i2c_adapter *adap;
	int curr_iic = 0;

	printk(KERN_INFO "iic_ibmocp_init: IBM on-chip iic adapter module\n");

    init_waitqueue_head(&iic_ibmocp_bus_waitq);
	
	while (curr_iic != -ENXIO) {
		if (!(iic_drv = ocp_alloc_dev(sizeof (struct i2c_adapter))))
			return -ENOMEM;
		iic_drv->type = IIC;
		if ((curr_iic = ocp_register(iic_drv)) != -ENXIO) {
			adap = (struct i2c_adapter *) iic_drv->ocpdev;
			iic_drv->vaddr = ioremap(iic_drv->paddr,17);
            fpga_base_addr = ioremap_nocache(FPGA_PHY_BASE_ADDR, 8);
			DEB2(printk
			    ("iic_hw_resrc_init: ioremapped base address: %p\n",
			     iic_drv->vaddr));
			DEB2(printk
			    ("Adapter irq %x\n",
			     iic_drv->irq));

			strcpy(adap->name, "IBM OCP IIC adapter");
			adap->data = (void *) iic_drv;
			adap->id = I2C_HW_OCP;
			adap->algo = NULL;
			adap->inc_use = iic_ibmocp_inc_use;
			adap->dec_use = iic_ibmocp_dec_use;
			adap->client_register = iic_ibmocp_reg;
			adap->client_unregister = iic_ibmocp_unreg;

			init_waitqueue_head(&(iic_wait[curr_iic]));
			init_waitqueue_head(&(slaveread_complete_iic_wait[curr_iic]));
			init_waitqueue_head(&(slaveread_service_iic_wait[curr_iic]));
			init_waitqueue_head(&(slavewrite_complete_iic_wait[curr_iic]));
			init_waitqueue_head(&(slavewrite_service_iic_wait[curr_iic]));

			if (iic_drv->irq > 0) {
				if (request_irq
				    (iic_drv->irq,
				     iic_ibmocp_handler, 0, "IBM OCP IIC",
				     iic_drv)) {
					printk(KERN_ERR "iic_hw_resrc_init: Request irq%d failed\n", iic_drv->irq);
					iic_drv->irq = 0;
				} else {
					DEB3(printk
					     ("iic_hw_resrc_init: Enabled interrupt\n"));
				}
			}

			if (i2c_iic_add_bus(adap) < 0)
				return -ENODEV;

			DEB2(printk
			    (KERN_INFO
			     "iic_ibmocp_init: found device at %p.\n\n",
			     iic_drv->vaddr));
		} else {
			ocp_free_dev(iic_drv);
			break;
		}		/* end if */
	}			/* end while */
	return (curr_iic == -ENXIO) ? 0 : curr_iic;
}

static void
iic_ibmocp_exit(void)
{
	int i;
	struct ocp_dev *iic_drv;
	struct i2c_adapter *adap;

	for (i = 0; i < ocp_get_max(IIC); i++) {
		iic_drv = ocp_get_dev(IIC, i);
		adap = (struct i2c_adapter *) iic_drv->ocpdev;
		i2c_iic_del_bus(adap);
	}

	iic_ibmocp_release();
}


/*
 * If modules is NOT defined when this file is compiled, then the MODULE_*
 * macros will resolve to nothing
 */
MODULE_AUTHOR("MontaVista Software <www.mvista.com>");
MODULE_DESCRIPTION("I2C-Bus adapter routines for IBM OCP IIC bus adapter");
MODULE_PARM(i2c_debug, "i");
MODULE_LICENSE("GPL");

module_init(iic_ibmocp_init);
module_exit(iic_ibmocp_exit);
