#include <ppcboot.h>
#include <command.h>

/* Use this #define to tweak the front ports to be IB */
/* plugfest compliant.  Pre-emphasis is not turned on */
/* for plugfest.                                      */
#if 0
#define PLUGFEST_COMPLIANT
#endif

#if (CONFIG_COMMANDS & CFG_CMD_ANAFA)

#include <asm/processor.h>

enum {
#if defined(CONFIG_TSIGRLT)
  NUM_ANAFA      = 3,
#elif defined(CONFIG_TS270SC)
  NUM_ANAFA      = 2,
#elif defined(CONFIG_TS120SC)
  NUM_ANAFA      = 1,
#elif defined(CONFIG_TSBLDSC)
  NUM_ANAFA      = 1,
#else
  NUM_ANAFA      = 6,
#endif
  ANAFA_NUM_PORT = 8,
  ANAFA_TIMEOUT  = 10000       /* default boot timeout in millisecs */
};

#if defined(CONFIG_TSIGRLT)
/* GPIO9  ->  ANAFA 1 (U1)
   GPIO10 ->  ANAFA 2 (U2)
   GPIO11 ->  ANAFA 3 (U3) */
#define anafa_rst_gpio_bitmask(anum) (1 << (23 - anum))
#elif defined(CONFIG_TS270SC)
/* GPIO9  ->  ANAFA 1 (U1)
   GPIO10 ->  ANAFA 2 (U2) */
#define anafa_rst_gpio_bitmask(anum) (1 << (23 - anum))
#elif defined(CONFIG_TS120SC)
/* GPIO9  ->  ANAFA 1 (U1) */
#define anafa_rst_gpio_bitmask(anum) (1 << (23 - anum))
#elif defined(CONFIG_TSBLDSC)
/* GPIO2  ->  ANAFA 1 (U1) */
#define anafa_rst_gpio_bitmask(anum) (1 << (30 - anum))
#else
/* GPIO1  ->  ANAFA 1 (U1)
   GPIO2  ->  ANAFA 2 (U2)
   GPIO3  ->  ANAFA 3 (U3)
   GPIO4  ->  ANAFA 4 (U4)
   GPIO5  ->  ANAFA 5 (U5)
   GPIO6  ->  ANAFA 6 (U6) */
#define anafa_rst_gpio_bitmask(anum) (1 << (31 - anum))
#endif

static const unsigned long anafa_base        = 0xf0200000;
static const unsigned long anafa_mem_size    = 0x10000;
static const unsigned long anafa_intctl      = 0x3010; /* offset of MISCC.INTCTL */
static const unsigned long anafa_boot_mask   = 0xc0000000; /* INITPCI and INITIB bits*/

static inline unsigned long anafa_read(int anafa, unsigned long offset) {
  return in32 (anafa_base + anafa_mem_size * (anafa - 1) + offset);
}

static inline void anafa_write(int anafa, unsigned long offset, unsigned long val) {
  out32 (anafa_base + anafa_mem_size * (anafa - 1) + offset, val);
}

int do_anafa (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
  int anafa;
  int anafamin, anafamax;
  int timeout;
  int value = 1;

  if (argc < 2 || argc > 4) {
    printf ("Usage:\n%s\n", cmdtp->usage);
    return 1;
  }

  if (argv[1][0] == 'b') {
    value = 1;
    ++argv;
    --argc;
  } else if (argv[1][0] == 'r') {
    value = 0;
    ++argv;
    --argc;
  }

  anafa = simple_strtoul(argv[1], NULL, 10);
  if (anafa < 0 || anafa > NUM_ANAFA) {
    printf ("Anafa number must be between 1 and %d, or 0 for all\n", NUM_ANAFA);
    return 1;
  }
  anafamax = anafa ? anafa : NUM_ANAFA;
  anafamin = anafa ? anafa : 1;

  timeout = argc == 3
    ? simple_strtoul(argv[2], NULL, 10)
    : ANAFA_TIMEOUT;

  for (anafa = anafamin ; anafa <= anafamax ; anafa++)
  {
    if (value) {
      printf ("Releasing Anafa %d from reset... \n", anafa);
    } else {
      printf ("Resetting Anafa %d.\n", anafa);
    }

    /* We want to drive a 1 on this GPIO pin to release an ANAFA from
       reset.  Per table 23-3 in the 405GP user manual, we need to set
       the corresponding bit in registers as:

       GPIO0_ODR --> 0
       GPIO0_OR  --> 1
       GPIO0_TCR --> 1

       To put the Anafa back into reset, we set

       GPIO0_OR  --> 0
    */
 
 
    out32(GPIO_ODR, in32(GPIO_ODR)  & ~anafa_rst_gpio_bitmask(anafa));
    out32(GPIO_TCR, in32(GPIO_TCR)  |  anafa_rst_gpio_bitmask(anafa));
    if (value) {
      out32(GPIO_OR,  in32(GPIO_OR) |  anafa_rst_gpio_bitmask(anafa));
    } else {
      out32(GPIO_OR,  in32(GPIO_OR) & ~anafa_rst_gpio_bitmask(anafa));
    }

    /* no waiting when we reset */
    if (!value) {
      continue;
    }
  }

  if (anafamin != anafamax) {
    udelay(8000000);
  }
  return 0;
}

/* ==========================
   Anafa SERDES configuration
   ========================== */

static const unsigned long anafa_phycon      = 0x8270; /* offset of IBPC[0].PHYCON */
static const unsigned long anafa_ibpc_size   = 0x400; /* size of IBPC entry */
static const unsigned long anafa_phycon_mask = 0xfcff0ff0;
static const unsigned long anafa_lpac        = 0x20000000;
static const unsigned long anafa_rlrac       = 0x40000000;
static const unsigned long anafa_tlrac       = 0x80000000;

static const unsigned long anafa_scd         = 0x8240; /* offset of IBPC[0].SCD */
static const unsigned long anafa_fullsw      = 0x0000f000;
static const unsigned long anafa_txemp       = 0x0000000f;
static const unsigned long anafa_eneql       = 0x0000f000;

enum tANAFA_PORT {
  ANAFA_PORT_INTERNAL,
  ANAFA_PORT_CONTROLLER,
  ANAFA_PORT_BACKPLANE,
  ANAFA_PORT_INTERSWITCH,
  ANAFA_PORT_FRONT
};

static const char *port_type_string[] = {
  [ANAFA_PORT_INTERNAL]    = "internal",
  [ANAFA_PORT_CONTROLLER]  = "controller",
  [ANAFA_PORT_BACKPLANE]   = "backplane",
  [ANAFA_PORT_INTERSWITCH] = "interswitch",
  [ANAFA_PORT_FRONT]       = "front"
};

struct tANAFA_STRUCT {
  enum tANAFA_PORT port[ANAFA_NUM_PORT];
};

/* IGR SC has 6 Anafas, IGR LT has 3 Anafas, 270 SC has 2 Anafas */
static struct tANAFA_STRUCT anafa_table[NUM_ANAFA] = {
  [0] = {
    .port = {
      [0] = ANAFA_PORT_CONTROLLER,
      [1] = ANAFA_PORT_BACKPLANE,
      [2] = ANAFA_PORT_BACKPLANE,
      [3] = ANAFA_PORT_INTERNAL,
      [4] = ANAFA_PORT_INTERNAL,
      [5] = ANAFA_PORT_INTERNAL,
      [6] = ANAFA_PORT_FRONT,
      [7] = ANAFA_PORT_FRONT,
    }
  },

  [1] = {
    .port = {
      [0] = ANAFA_PORT_BACKPLANE,
      [1] = ANAFA_PORT_BACKPLANE,
      [2] = ANAFA_PORT_INTERNAL,
      [3] = ANAFA_PORT_INTERNAL,
      [4] = ANAFA_PORT_INTERNAL,
      [5] = ANAFA_PORT_FRONT,
      [6] = ANAFA_PORT_FRONT,
      [7] = ANAFA_PORT_INTERNAL,
    }
  },

#if defined(CONFIG_TSIGRSC) || defined(CONFIG_TSIGRLT)
  [2] = {
    .port = {
      [0] = ANAFA_PORT_INTERSWITCH,
      [1] = ANAFA_PORT_BACKPLANE,
      [2] = ANAFA_PORT_BACKPLANE,
      [3] = ANAFA_PORT_INTERNAL,
      [4] = ANAFA_PORT_INTERNAL,
      [5] = ANAFA_PORT_FRONT,
      [6] = ANAFA_PORT_FRONT,
      [7] = ANAFA_PORT_INTERNAL,
    }
  },

#if defined(CONFIG_TSIGRSC)
  [3] = {
    .port = {
      [0] = ANAFA_PORT_INTERNAL,
      [1] = ANAFA_PORT_BACKPLANE,
      [2] = ANAFA_PORT_BACKPLANE,
      [3] = ANAFA_PORT_INTERSWITCH,
      [4] = ANAFA_PORT_INTERNAL,
      [5] = ANAFA_PORT_FRONT,
      [6] = ANAFA_PORT_FRONT,
      [7] = ANAFA_PORT_INTERNAL,
    }
  },

  [4] = {
    .port = {
      [0] = ANAFA_PORT_INTERNAL,
      [1] = ANAFA_PORT_INTERNAL,
      [2] = ANAFA_PORT_BACKPLANE,
      [3] = ANAFA_PORT_BACKPLANE,
      [4] = ANAFA_PORT_INTERNAL,
      [5] = ANAFA_PORT_FRONT,
      [6] = ANAFA_PORT_FRONT,
      [7] = ANAFA_PORT_INTERNAL,
    }
  },

  [5] = {
    .port = {
      [0] = ANAFA_PORT_INTERNAL,
      [1] = ANAFA_PORT_BACKPLANE,
      [2] = ANAFA_PORT_BACKPLANE,
      [3] = ANAFA_PORT_CONTROLLER,
      [4] = ANAFA_PORT_FRONT,
      [5] = ANAFA_PORT_FRONT,
      [6] = ANAFA_PORT_INTERNAL,
      [7] = ANAFA_PORT_INTERNAL,
    }
  }
#endif /* TSIGRSC */
#endif /* TSIGRLT || TSIGRSC */
};

int do_serdes_cfg (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[]) {
  int a, p;

  printf ("Configuring SERDES:\n");

  for (a = 0; a < NUM_ANAFA; ++a) {
    printf ("  Anafa %d\n", a + 1);

    for (p = 0; p < ANAFA_NUM_PORT; ++p) {
      long scd;

      printf ("    Port %d: %s ...", p + 1, port_type_string[anafa_table[a].port[p]]);

      /* turn on RX equalization for all ports */
      scd  = anafa_read(a + 1, anafa_scd + p * anafa_ibpc_size + 4);
      scd |= anafa_eneql;
      anafa_write(a + 1, anafa_scd + p * anafa_ibpc_size + 4, scd);
      udelay(1000);

      /* port auto configuration for all ports */
      {
        long phycon;

        /* set lane polarity to 0 and clear lane reversal */
        phycon  = anafa_read(a + 1, anafa_phycon + p * anafa_ibpc_size);
        phycon &= anafa_phycon_mask;
        anafa_write(a + 1, anafa_phycon + p * anafa_ibpc_size, phycon);
        udelay(1000);

        /* turn on auto-configure */
        phycon  = anafa_read(a + 1, anafa_phycon + p * anafa_ibpc_size);
        phycon |= anafa_lpac;
        anafa_write(a + 1, anafa_phycon + p * anafa_ibpc_size, phycon);
        udelay(1000);

        phycon  = anafa_read(a + 1, anafa_phycon + p * anafa_ibpc_size);
        phycon |= anafa_rlrac;
        anafa_write(a + 1, anafa_phycon + p * anafa_ibpc_size, phycon);
        udelay(1000);

/* There was a problem when both sides turn on RxLaneRevAutoConfig and TxLaneRevAutoConfig.
 * This problem showed up in manufacturing test where the link between Anafa 1 port 6 and Anafa 6 port 6 stuck in traning state.
 * Korkut suggested to turn off TxLaneRevAutoConfig and only use RxLaneRevAutoConfig.
 */
/* Disabling the code fragment bellow breaks the 1x link to the CC on TS360 */
/* Ok, we will still disable the code below as a default for all ports. For ports connecting to the controller, we will turn the bit back on for the link to come up */
/* Re-enabled - disabling this code fragment prevents the FIG IB links to come up physically */
/* Ok, let's turn on Tx Auto Lane Reversal for links to the controllers and to the gateway (backplane). Internal links and interswitch links should have this bit off.  We do this in the case statement below */
 
#if 0 
        phycon  = anafa_read(a + 1, anafa_phycon + p * anafa_ibpc_size);
        phycon |= anafa_tlrac;
        anafa_write(a + 1, anafa_phycon + p * anafa_ibpc_size, phycon);
        udelay(1000);
#endif

      }

      switch (anafa_table[a].port[p]) {
      case ANAFA_PORT_INTERNAL:
        /* set half-swing, no preemphasis */
        scd  = anafa_read(a + 1, anafa_scd + p * anafa_ibpc_size);
        scd &= ~anafa_fullsw;
        anafa_write(a + 1, anafa_scd + p * anafa_ibpc_size, scd);
        udelay(1000);

        scd  = anafa_read(a + 1, anafa_scd + p * anafa_ibpc_size + 4);
        scd &= ~anafa_txemp;
        anafa_write(a + 1, anafa_scd + p * anafa_ibpc_size + 4, scd);
        udelay(1000);

        break;

      case ANAFA_PORT_CONTROLLER:

        /* fall through to backplane link settings: */

      case ANAFA_PORT_BACKPLANE:

/* Turn on the Tx Lane Reversal Auto Config bit for ports connecting to the controller */
        {
          long phycon;

          phycon  = anafa_read(a + 1, anafa_phycon + p * anafa_ibpc_size);
          phycon |= anafa_tlrac;
          anafa_write(a + 1, anafa_phycon + p * anafa_ibpc_size, phycon);
          udelay(1000);
        }

      case ANAFA_PORT_INTERSWITCH:
        /* set full-swing with preemphasis */
        scd  = anafa_read(a + 1, anafa_scd + p * anafa_ibpc_size);
        scd |= anafa_fullsw;
        anafa_write(a + 1, anafa_scd + p * anafa_ibpc_size, scd);
        udelay(1000);

        scd  = anafa_read(a + 1, anafa_scd + p * anafa_ibpc_size + 4);
        scd |= anafa_txemp;
        anafa_write(a + 1, anafa_scd + p * anafa_ibpc_size + 4, scd);
        udelay(1000);

        break;

      case ANAFA_PORT_FRONT:
        /* set full-swing */
        scd  = anafa_read(a + 1, anafa_scd + p * anafa_ibpc_size);
        scd |= anafa_fullsw;
        anafa_write(a + 1, anafa_scd + p * anafa_ibpc_size, scd);
        udelay(1000);

#ifndef PLUGFEST_COMPLIANT
        /* set preemphasis */
        scd  = anafa_read(a + 1, anafa_scd + p * anafa_ibpc_size + 4);
        scd |= anafa_txemp;
        anafa_write(a + 1, anafa_scd + p * anafa_ibpc_size + 4, scd);
        udelay(1000);
#endif

        break;
      }

      printf (" done\n");
    }
  }

  return 0;
}

#endif	/* CFG_CMD_ANAFA */
